#
# SPDX-FileCopyrightText: Copyright (c) 2021-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
from hashlib import sha384
import os
from enum import Enum
import logging
import sys
from collections import defaultdict
from verifier.__about__ import __author__, __copyright__, __version__

info_log = logging.getLogger('gpu-verifier-info')
info_log.setLevel(logging.INFO)
shandler = logging.StreamHandler(sys.stdout)
info_log.addHandler(shandler)

parent_dir = os.path.dirname(os.path.abspath(__file__))
logger_file_path = os.path.join(os.getcwd(), "verifier.log")

if os.path.exists(logger_file_path):
    os.remove(logger_file_path)

event_log = logging.getLogger('gpu-verifier-event')
event_log.setLevel(logging.DEBUG)
fhandler = logging.FileHandler(logger_file_path)
fhandler.setFormatter(logging.Formatter("%(asctime)s:%(levelname)s: %(message)s", '%m-%d-%Y %H:%M:%S'))
event_log.addHandler(fhandler)

event_log.info(f'Verifier version: {__version__}')
event_log.debug("----------STARTING----------")


class BaseSettings:
    SIZE_OF_NONCE_IN_BYTES = 32
    SIZE_OF_NONCE_IN_HEX_STR = 64
    gpu_availability = False
    attestation_report_availability = False
    TEST_NO_GPU_NUMBER_OF_GPUS = 1
    NONCE = "4cff7f5380ead8fad8ec2c531c110aca4302a88f603792801a8ca29ee151af2e"
    # The maximum number of times the CC ADMIN will retry the GPU attestation.
    MAX_RETRY_COUNT = 3
    current_retry_count = 0
    # The maximum time in seconds to wait for an NVML call before timing out.
    MAX_NVML_TIME_DELAY = 30
    MAX_OCSP_TIME_DELAY = 10
    MAX_NETWORK_TIME_DELAY = 10
    OCSP_URL = os.getenv('NV_OCSP_URL', 'https://ocsp.ndis.nvidia.com/')
    OCSP_NONCE_DISABLED = False
    CLAIMS_VERSION = "2.0"
    OCSP_HASH_FUNCTION = sha384
    RIM_SERVICE_BASE_URL = os.getenv('NV_RIM_URL', 'https://rim.attestation.nvidia.com/v1/rim/')
    Certificate_Chain_Verification_Mode = Enum("CERT CHAIN VERIFICATION MODE",
                                               ['GPU_ATTESTATION', 'OCSP_RESPONSE', 'DRIVER_RIM_CERT',
                                                'VBIOS_RIM_CERT'])
    NVDEC_STATUS = Enum("NVDEC0 status", [("ENABLED", 0xAA), ("DISABLED", 0x55)])
    class Status(Enum):
        VALID = "valid"
        INVALID = "invalid"
        REVOKED = "revoked"
        EXPIRED = "expired"
    INDEX_OF_IK_CERT = 1
    SKU = "PROD"
    claims = {}
    allow_hold_cert = False
    service_key = None
    ROOT_CERT_DIR = os.path.join(parent_dir, "certs")
    RIM_ROOT_CERT = os.path.join(ROOT_CERT_DIR, 'verifier_RIM_root.pem')
    DEVICE_ROOT_CERT = os.path.join(ROOT_CERT_DIR, 'verifier_device_root.pem')

    EXECUTION_SEQUENCE_INDEX = {
        'GPU_AVAILABILITY'                       : 0,
        'ATTESTATION_REPORT_AVAILABILITY'        : 1,
        'GPU_INFO_FETCH'                         : 2,
        'CORRECT_GPU_ARCH'                       : 3,
        'ATTESTATION_REPORT_MSR_AVAILABILITY'    : 4,
        'ATTESTATION_REPORT_PARSED'              : 5,
        'ROOT_CERT_AVAILABILITY'                 : 6,
        'GPU_CERT_CHAIN_VERIFIED'                : 7,
        'GPU_CERT_OCSP_CERT_CHAIN_VERIFICATION'  : 8,
        'GPU_CERT_OCSP_SIGNATURE_VERIFICATION'   : 9,
        'GPU_CERT_OCSP_NONCE_MATCH'              : 10,
        'GPU_CERT_CHECK_COMPLETE'                : 11,
        'NONCE_MATCH'                            : 12,
        'ATTESTATION_REPORT_DRV_VERSION_MATCH'   : 13,
        'ATTESTATION_REPORT_VBIOS_VERSION_MATCH' : 14,
        'ATTESTATION_REPORT_VERIFICATION'        : 15,
        'DRIVER_RIM_FETCH'                       : 16,
        'DRIVER_RIM_MEASUREMENT_AVAILABILITY'    : 17,
        'DRIVER_RIM_SCHEMA_VALIDATION'           : 18,
        'DRIVER_RIM_VERSION_MATCH'               : 19,
        'DRIVER_RIM_CERT_EXTRACT'                : 20,
        'DRIVER_RIM_SIGNATURE_VERIFICATION'      : 21,
        'VBIOS_RIM_FETCH'                        : 22,
        'VBIOS_RIM_MEASUREMENT_AVAILABILITY'     : 23,
        'VBIOS_RIM_SCHEMA_VALIDATION'            : 24,
        'VBIOS_RIM_VERSION_MATCH'                : 25,
        'VBIOS_RIM_CERT_EXTRACT'                 : 26,
        'VBIOS_RIM_SIGNATURE_VERIFICATION'       : 27,
        'DRV_VBIOS_MSR_INDEX_CONFLICT'           : 28,
        'MEASUREMENT_MATCH'                      : 29,
    }

    CURRENT_OPAQUE_DATA_VERSION = 0
    LATEST_OPAQUE_DATA_VERSION = 1
    MIN_OPAQUE_DATA_VERSION_TO_SUPPORT_FEATURE_FLAG = 1
    MIN_OPAQUE_DATA_VERSION_TO_SUPPORT_CHIP_INFO = 1

    @classmethod
    def set_ocsp_url(cls, url):
        if not isinstance(url, str):
            raise ValueError("Incorrect data type for the URL.")
        if not url:
            raise ValueError("OCSP URL is empty")
        if not url.endswith('/'):
            url += '/'
        cls.OCSP_URL = url

    @classmethod
    def set_rim_service_base_url(cls, url):
        if not isinstance(url, str):
            raise ValueError("Incorrect data type for the URL.")
        if not url:
            raise ValueError("RIM URL is empty")
        if not url.endswith('/'):
            url += '/'
        cls.RIM_SERVICE_BASE_URL = url

    @classmethod
    def set_service_key(cls, service_key):
        if not isinstance(service_key, str):
            raise ValueError("Incorrect data type for the service_key.")
        cls.service_key = service_key

    @classmethod
    def get_sku(cls):
        return cls.SKU

    @classmethod
    def set_sku(cls, sku):
        cls.SKU = sku

    @classmethod
    def is_retry_allowed(cls):
        if (cls.current_retry_count < cls.MAX_RETRY_COUNT):
            cls.current_retry_count += 1
            return True
        else:
            return False

    @classmethod
    def reset(cls):
        cls.NONCE = bytes.fromhex("4cff7f5380ead8fad8ec2c531c110aca4302a88f603792801a8ca29ee151af2e")
        cls.current_retry_count = 0
        cls.claims = {}

    @classmethod
    def set_nonce(cls, nonce):
        cls.NONCE = nonce

    def __init__(self):
        self.measurement_comparison = None
        self.gpu_arch_is_correct = None
        self.parse_attestation_report = None
        self.nonce_comparison = None
        self.attestation_report_driver_version_match = None
        self.attestation_report_vbios_version_match = None
        self.gpu_driver_rim_version_matched = None
        self.gpu_vbios_rim_version_matched = None
        self.rim_driver_measurements_availability = None
        self.rim_vbios_measurements_availability = None
        self.gpu_driver_rim_schema_validated = None
        self.gpu_vbios_rim_schema_validated = None
        self.gpu_driver_rim_signature_verified = None
        self.gpu_vbios_rim_signature_verified = None
        self.fetch_driver_rim = None
        self.fetch_vbios_rim = None
        self.no_driver_vbios_measurement_index_conflict = None
        self.gpu_attestation_report_cert_chain_validated = None
        self.attestation_report_signature_verification = None
        self.gpu_driver_version = None
        self.gpu_vbios_version = None
        self.gpu_attestation_report_cert_chain_fwid_matched = None
        self.gpu_driver_rim_cert_chain_validated = None
        self.gpu_vbios_rim_cert_chain_validated = None
        self.gpu_attestation_report_cert_expiration_date = None
        self.gpu_driver_rim_cert_expiration_date = None
        self.gpu_vbios_rim_cert_expiration_date = None
        self.gpu_attestation_report_cert_revocation_reason = None
        self.gpu_driver_rim_cert_revocation_reason = None
        self.gpu_vbios_rim_cert_revocation_reason = None
        self.gpu_attestation_report_cert_ocsp_status = None
        self.gpu_driver_rim_cert_ocsp_status = None
        self.gpu_vbios_rim_cert_ocsp_status = None
        self.gpu_attestation_report_cert_status = None
        self.gpu_driver_rim_cert_status = None
        self.gpu_vbios_rim_cert_status = None

    @classmethod
    def mark_attestation_report_as_available(cls, flag=True):
        event_log.debug("mark_attestation_report_as_available called.")
        cls.attestation_report_availability = flag

    def check_if_gpu_attestation_report_cert_chain_validated(self):
        event_log.debug(
            f"check_if_gpu_attestation_report_cert_chain_validated: {self.gpu_attestation_report_cert_chain_validated}")
        return self.gpu_attestation_report_cert_chain_validated

    def mark_gpu_attestation_report_cert_chain_validated(self, flag=True):
        event_log.debug("mark_gpu_attestation_report_cert_chain_validated called")
        self.gpu_attestation_report_cert_chain_validated = flag

    def check_gpu_attestation_report_cert_status(self):
        event_log.debug(f"check_gpu_attestation_report_cert_status called.{self.gpu_attestation_report_cert_status}")
        return self.gpu_attestation_report_cert_status

    def mark_gpu_attestation_report_cert_status(self, cert_status):
        event_log.debug("mark_gpu_attestation_report_cert_status called.")
        self.gpu_attestation_report_cert_status = cert_status

    def check_gpu_driver_rim_cert_status(self):
        event_log.debug(f"check_gpu_driver_rim_cert_status called.{self.gpu_driver_rim_cert_status}")
        return self.gpu_driver_rim_cert_status

    def mark_gpu_driver_rim_cert_status(self, cert_status):
        event_log.debug("mark_gpu_driver_rim_cert_status called.")
        self.gpu_driver_rim_cert_status = cert_status

    def check_gpu_vbios_rim_cert_status(self):
        event_log.debug(f"check_gpu_vbios_rim_cert_status called.{self.gpu_vbios_rim_cert_status}")
        return self.gpu_vbios_rim_cert_status

    def mark_gpu_vbios_rim_cert_status(self, cert_status):
        event_log.debug("mark_gpu_vbios_rim_cert_status called.")
        self.gpu_vbios_rim_cert_status = cert_status

    def check_gpu_attestation_report_cert_revocation_reason(self):
        event_log.debug(
            f"check_gpu_attestation_report_cert_revocation_reason called.{self.gpu_attestation_report_cert_revocation_reason}")
        return self.gpu_attestation_report_cert_revocation_reason

    def mark_gpu_attestation_report_cert_revocation_reason(self, revocation_reason):
        event_log.debug("mark_gpu_attestation_report_cert_revocation_reason called.")
        self.gpu_attestation_report_cert_revocation_reason = revocation_reason

    def check_gpu_driver_rim_cert_revocation_reason(self):
        event_log.debug(
            f"check_gpu_driver_rim_cert_revocation_reason called.{self.gpu_driver_rim_cert_revocation_reason}")
        return self.gpu_driver_rim_cert_revocation_reason

    def mark_gpu_driver_rim_cert_revocation_reason(self, revocation_reason):
        event_log.debug("mark_gpu_driver_rim_cert_revocation_reason called.")
        self.gpu_driver_rim_cert_revocation_reason = revocation_reason

    def check_gpu_vbios_rim_cert_revocation_reason(self):
        event_log.debug(
            f"check_gpu_vbios_rim_cert_revocation_reason called.{self.gpu_vbios_rim_cert_revocation_reason}")
        return self.gpu_vbios_rim_cert_revocation_reason

    def mark_gpu_vbios_rim_cert_revocation_reason(self, revocation_reason):
        event_log.debug("mark_gpu_vbios_rim_cert_revocation_reason called.")
        self.gpu_vbios_rim_cert_revocation_reason = revocation_reason

    def check_gpu_attestation_report_cert_ocsp_status(self):
        event_log.debug(
            f"check_gpu_attestation_report_cert_ocsp_status: {self.gpu_attestation_report_cert_ocsp_status}")
        return self.gpu_attestation_report_cert_ocsp_status

    def mark_gpu_attestation_report_cert_ocsp_status(self, ocsp_status):
        event_log.debug("mark_gpu_attestation_report_cert_ocsp_status called.")
        self.gpu_attestation_report_cert_ocsp_status = ocsp_status

    def check_gpu_driver_rim_cert_ocsp_status(self):
        event_log.debug(f"check_gpu_driver_rim_cert_ocsp_status: {self.gpu_driver_rim_cert_ocsp_status}")
        return self.gpu_driver_rim_cert_ocsp_status

    def mark_gpu_driver_rim_cert_ocsp_status(self, ocsp_status):
        event_log.debug("mark_gpu_driver_rim_cert_ocsp_status called.")
        self.gpu_driver_rim_cert_ocsp_status = ocsp_status

    def check_gpu_vbios_rim_cert_ocsp_status(self):
        event_log.debug(f"check_gpu_vbios_rim_cert_ocsp_status: {self.gpu_vbios_rim_cert_ocsp_status}")
        return self.gpu_vbios_rim_cert_ocsp_status

    def mark_gpu_vbios_rim_cert_ocsp_status(self, ocsp_status):
        event_log.debug("mark_gpu_vbios_rim_cert_ocsp_status called.")
        self.gpu_vbios_rim_cert_ocsp_status = ocsp_status

    def check_if_gpu_driver_rim_cert_chain_validated(self):
        event_log.debug(f"check_if_gpu_driver_rim_cert_chain_validated: {self.gpu_driver_rim_cert_chain_validated}")
        return self.gpu_driver_rim_cert_chain_validated

    def mark_gpu_driver_rim_cert_chain_validated(self, flag=True):
        event_log.debug("mark_gpu_driver_rim_cert_chain_validated called")
        self.gpu_driver_rim_cert_chain_validated = flag

    def check_if_gpu_vbios_rim_cert_chain_validated(self):
        event_log.debug(f"check_if_gpu_vbios_rim_cert_chain_validated: {self.gpu_vbios_rim_cert_chain_validated}")
        return self.gpu_vbios_rim_cert_chain_validated

    def mark_gpu_vbios_rim_cert_chain_validated(self, flag=True):
        event_log.debug("mark_gpu_vbios_rim_cert_chain_validated called")
        self.gpu_vbios_rim_cert_chain_validated = flag

    def check_gpu_attestation_report_cert_expiration_date(self):
        event_log.debug(
            f"check_gpu_attestation_report_cert_expiration_date called.{self.gpu_attestation_report_cert_expiration_date}")
        return self.gpu_attestation_report_cert_expiration_date

    def mark_gpu_attestation_report_cert_expiration_date(self, expiration_date):
        event_log.debug("mark_gpu_attestation_report_cert_expiration_date called.")
        self.gpu_attestation_report_cert_expiration_date = expiration_date

    def check_gpu_driver_rim_cert_expiration_date(self):
        event_log.debug(f"check_gpu_driver_rim_cert_expiration_date called.{self.gpu_driver_rim_cert_expiration_date}")
        return self.gpu_driver_rim_cert_expiration_date

    def mark_gpu_driver_rim_cert_expiration_date(self, expiration_date):
        event_log.debug("mark_gpu_driver_rim_cert_expiration_date called.")
        self.gpu_driver_rim_cert_expiration_date = expiration_date

    def check_gpu_vbios_rim_cert_expiration_date(self):
        event_log.debug(f"check_gpu_vbios_rim_cert_expiration_date called.{self.gpu_vbios_rim_cert_expiration_date}")
        return self.gpu_vbios_rim_cert_expiration_date

    def mark_gpu_vbios_rim_cert_expiration_date(self, expiration_date):
        event_log.debug("mark_gpu_vbios_rim_cert_expiration_date called.")
        self.gpu_vbios_rim_cert_expiration_date = expiration_date

    def check_if_gpu_attestation_report_cert_chain_fwid_matched(self):
        event_log.debug(
            f"check_if_gpu_attestation_report_cert_chain_fwid_matched: {self.gpu_attestation_report_cert_chain_fwid_matched}")
        return self.gpu_attestation_report_cert_chain_fwid_matched

    def mark_gpu_attestation_report_cert_chain_fwid_matched(self, flag=True):
        event_log.debug("mark_gpu_attestation_report_cert_chain_fwid_matched called")
        self.gpu_attestation_report_cert_chain_fwid_matched = flag

    def check_if_attestation_report_signature_verified(self):
        event_log.debug(
            f"check_if_attestation_report_signature_verified: {self.attestation_report_signature_verification}")
        return self.attestation_report_signature_verification

    def mark_attestation_report_signature_verified(self, flag=True):
        event_log.debug("mark_attestation_report_signature_verified called")
        self.attestation_report_signature_verification = flag

    def check_if_driver_rim_fetched(self):
        event_log.debug(f"check_if_driver_rim_fetched: {self.fetch_driver_rim}")
        return self.fetch_driver_rim

    def mark_driver_rim_fetched(self):
        event_log.debug("mark_driver_rim_fetched called")
        self.fetch_driver_rim = True

    def check_if_vbios_rim_fetched(self):
        event_log.debug(f"check_if_vbios_rim_fetched: {self.fetch_vbios_rim}")
        return self.fetch_vbios_rim

    def mark_vbios_rim_fetched(self):
        event_log.debug("mark_vbios_rim_fetched called.")
        self.fetch_vbios_rim = True

    def check_if_gpu_driver_rim_signature_verified(self):
        event_log.debug(f"check_if_gpu_driver_rim_signature_verified: {self.gpu_driver_rim_signature_verified}")
        return self.gpu_driver_rim_signature_verified

    def mark_gpu_driver_rim_signature_verified(self, flag=True):
        event_log.debug("mark_gpu_driver_rim_signature_verified called.")
        self.gpu_driver_rim_signature_verified = flag

    def check_if_gpu_vbios_rim_signature_verified(self):
        event_log.debug(f"check_if_gpu_vbios_rim_signature_verified: {self.gpu_vbios_rim_signature_verified}")
        return self.gpu_vbios_rim_signature_verified

    def mark_gpu_vbios_rim_signature_verified(self, flag=True):
        event_log.debug("mark_gpu_vbios_rim_signature_verified called.")
        self.gpu_vbios_rim_signature_verified = flag

    def check_if_gpu_driver_rim_schema_validated(self):
        event_log.debug(f"check_if_gpu_driver_rim_schema_validated: {self.gpu_driver_rim_schema_validated}")
        return self.gpu_driver_rim_schema_validated

    def mark_gpu_driver_rim_schema_validated(self, flag=True):
        event_log.debug("mark_gpu_driver_rim_schema_validated called.")
        self.gpu_driver_rim_schema_validated = flag

    def check_gpu_driver_version(self):
        event_log.debug(f"check_gpu_driver_version called.{self.gpu_driver_version}")
        return self.gpu_driver_version

    def mark_gpu_driver_version(self, driver_version):
        event_log.debug("mark_gpu_driver_version called.")
        self.gpu_driver_version = driver_version

    def check_gpu_vbios_version(self):
        event_log.debug(f"check_gpu_vbios_version called.{self.gpu_vbios_version}")
        return self.gpu_vbios_version

    def mark_gpu_vbios_version(self, vbios_version):
        event_log.debug("mark_gpu_vbios_version called.")
        if vbios_version is not None:
            self.gpu_vbios_version = vbios_version.upper()

    def check_if_gpu_vbios_rim_schema_validated(self):
        event_log.debug(f"check_if_gpu_vbios_rim_schema_validated: {self.gpu_vbios_rim_schema_validated}")
        return self.gpu_vbios_rim_schema_validated

    def mark_gpu_vbios_rim_schema_validated(self, flag=True):
        event_log.debug("mark_gpu_vbios_rim_schema_validated called.")
        self.gpu_vbios_rim_schema_validated = flag

    def check_rim_driver_measurements_availability(self):
        event_log.debug(f"check_rim_driver_measurements_availability: {self.rim_driver_measurements_availability}")
        return self.rim_driver_measurements_availability

    def mark_rim_driver_measurements_as_available(self, flag=True):
        event_log.debug("mark_rim_driver_measurements_as_available called.")
        self.rim_driver_measurements_availability = flag

    def check_rim_vbios_measurements_availability(self):
        event_log.debug(f"check_rim_vbios_measurements_availability: {self.rim_vbios_measurements_availability}")
        return self.rim_vbios_measurements_availability

    def mark_rim_vbios_measurements_as_available(self, flag=True):
        event_log.debug("mark_rim_vbios_measurements_as_available called.")
        self.rim_vbios_measurements_availability = flag

    def check_if_measurements_are_matching(self):
        if self.measurement_comparison is True:
            return "success"
        elif self.measurement_comparison is False:
            return "fail"
        else:
            return None

    def mark_measurements_as_matching(self, flag=True):
        event_log.debug("mark_measurements_as_matching called.")
        self.measurement_comparison = flag

    def check_if_gpu_driver_rim_version_matched(self):
        event_log.debug(f"check_if_gpu_driver_rim_version_matched: {self.gpu_driver_rim_version_matched}")
        return self.gpu_driver_rim_version_matched

    def mark_gpu_driver_rim_version_matched(self, flag=True):
        event_log.debug("mark_gpu_driver_rim_version_matched called.")
        self.gpu_driver_rim_version_matched = flag

    def check_if_gpu_vbios_rim_version_matched(self):
        event_log.debug(f"check_if_gpu_vbios_rim_version_matched: {self.gpu_vbios_rim_version_matched}")
        return self.gpu_vbios_rim_version_matched

    def mark_gpu_vbios_rim_version_matched(self, flag=True):
        event_log.debug("mark_gpu_vbios_rim_version_matched called.")
        self.gpu_vbios_rim_version_matched = flag

    def check_if_gpu_arch_is_correct(self):
        event_log.debug(f"check_if_gpu_arch_is_correct: {self.gpu_arch_is_correct}")
        return self.gpu_arch_is_correct

    def mark_gpu_arch_is_correct(self, flag=True):
        event_log.debug("mark_gpu_arch_is_correct called.")
        self.gpu_arch_is_correct = flag

    def check_if_nonce_are_matching(self):
        event_log.debug(f"check_if_nonce_are_matching: {self.nonce_comparison}")
        return self.nonce_comparison

    def mark_nonce_as_matching(self, flag=True):
        event_log.debug("mark_nonce_as_matching called.")
        self.nonce_comparison = flag

    def check_if_attestation_report_parsed_successfully(self):
        event_log.debug(f"check_if_attestation_report_parsed_successfully: {self.parse_attestation_report}")
        return self.parse_attestation_report

    def mark_attestation_report_parsed(self):
        event_log.debug("mark_attestation_report_parsed called.")
        self.parse_attestation_report = True

    def check_if_attestation_report_driver_version_matches(self):
        event_log.debug(
            f"check_if_attestation_report_driver_version_matches: {self.attestation_report_driver_version_match}")
        return self.attestation_report_driver_version_match

    def mark_attestation_report_driver_version_as_matching(self, flag=True):
        event_log.debug("mark_attestation_report_driver_version_as_matching called.")
        self.attestation_report_driver_version_match = flag

    def check_if_attestation_report_vbios_version_matches(self):
        event_log.debug(
            f"check_if_attestation_report_vbios_version_matches: {self.attestation_report_vbios_version_match}")
        return self.attestation_report_vbios_version_match

    def mark_attestation_report_vbios_version_as_matching(self, flag=True):
        event_log.debug("mark_attestation_report_vbios_version_as_matching called.")
        self.attestation_report_vbios_version_match = flag

    def check_if_no_driver_vbios_measurement_index_conflict(self):
        event_log.debug(
            f"check_if_no_driver_vbios_measurement_index_conflict: {self.no_driver_vbios_measurement_index_conflict}")
        return self.no_driver_vbios_measurement_index_conflict

    def mark_no_driver_vbios_measurement_index_conflict(self, flag=True):
        event_log.debug("mark_no_driver_vbios_measurement_conflict called.")
        self.no_driver_vbios_measurement_index_conflict = flag

    def check_status(self):
        if self.check_if_gpu_arch_is_correct() and \
                self.check_if_gpu_attestation_report_cert_chain_validated() and \
                self.check_if_attestation_report_parsed_successfully() and \
                self.check_if_nonce_are_matching() and \
                ((self.check_gpu_attestation_report_cert_status() == 'valid') or
                 (
                         BaseSettings.allow_hold_cert and self.check_gpu_attestation_report_cert_status() == 'revoked' and self.check_gpu_attestation_report_cert_revocation_reason() == 'certificate_hold')) and \
                ((self.check_gpu_attestation_report_cert_ocsp_status() == 'good') or
                 (
                         BaseSettings.allow_hold_cert and self.check_gpu_attestation_report_cert_ocsp_status() == 'revoked' and self.check_gpu_attestation_report_cert_revocation_reason() == 'certificate_hold')) and \
                self.check_if_attestation_report_driver_version_matches() and \
                self.check_if_attestation_report_vbios_version_matches() and \
                self.check_if_attestation_report_signature_verified() and \
                self.check_if_gpu_attestation_report_cert_chain_fwid_matched and \
                self.check_if_driver_rim_fetched() and \
                self.check_if_gpu_driver_rim_schema_validated() and \
                self.check_if_gpu_driver_rim_cert_chain_validated() and \
                ((self.check_gpu_driver_rim_cert_status() == 'valid') or
                 (
                         BaseSettings.allow_hold_cert and self.check_gpu_driver_rim_cert_status() == 'revoked' and self.check_gpu_driver_rim_cert_revocation_reason() == 'certificate_hold')) and \
                ((self.check_gpu_driver_rim_cert_ocsp_status() == 'good') or
                 (
                         BaseSettings.allow_hold_cert and self.check_gpu_driver_rim_cert_ocsp_status() == 'revoked' and self.check_gpu_driver_rim_cert_revocation_reason() == 'certificate_hold')) and \
                self.check_if_gpu_driver_rim_signature_verified() and \
                self.check_if_gpu_driver_rim_version_matched() and \
                self.check_rim_driver_measurements_availability() and \
                self.check_if_vbios_rim_fetched() and \
                self.check_if_gpu_vbios_rim_schema_validated() and \
                self.check_if_gpu_vbios_rim_signature_verified() and \
                self.check_if_gpu_vbios_rim_version_matched() and \
                self.check_if_gpu_vbios_rim_cert_chain_validated() and \
                ((self.check_gpu_vbios_rim_cert_status() == 'valid') or
                 (
                         BaseSettings.allow_hold_cert and self.check_gpu_vbios_rim_cert_status() == 'revoked' and self.check_gpu_vbios_rim_cert_revocation_reason() == 'certificate_hold')) and \
                ((self.check_gpu_vbios_rim_cert_ocsp_status() == 'good') or
                 (
                         BaseSettings.allow_hold_cert and self.check_gpu_vbios_rim_cert_ocsp_status() == 'revoked' and self.check_gpu_vbios_rim_cert_revocation_reason() == 'certificate_hold')) and \
                self.check_rim_vbios_measurements_availability() and \
                self.check_if_no_driver_vbios_measurement_index_conflict() and \
                self.check_if_measurements_are_matching() == "success":
            BaseSettings.test_result = True
            return True
        else:
            BaseSettings.test_result = False
            return False


class HopperSettings(BaseSettings):
    signature_length = 96
    HashFunction = sha384
    MAX_CERT_CHAIN_LENGTH = 5
    HashFunctionNamespace = "{http://www.w3.org/2001/04/xmlenc#sha384}"
    GPU_ARCH_NAME = "HOPPER"
    GPU_ARCH = 9
    RIM_DIRECTORY_PATH = os.path.join(parent_dir, "samples")
    TEST_NO_GPU_DRIVER_RIM_PATH = os.path.join(RIM_DIRECTORY_PATH, "Driver_RIM_test_no_gpu.swidtag")
    DRIVER_RIM_PATH = ""
    TEST_NO_GPU_VBIOS_RIM_PATH = os.path.join(RIM_DIRECTORY_PATH, "1010_0200_882_96005E0001_test_no_gpu.swidtag")
    VBIOS_RIM_PATH = ""
    ATTESTATION_REPORT_PATH = os.path.join(RIM_DIRECTORY_PATH, "hopperAttestationReport.txt")
    GPU_ATTESTATION_CERTIFICATES_PATH = os.path.join(RIM_DIRECTORY_PATH, "hopperCertChain.txt")

    @classmethod
    def set_driver_rim_path(cls, path):
        cls.DRIVER_RIM_PATH = path

    @classmethod
    def set_vbios_rim_path(cls, path):
        cls.VBIOS_RIM_PATH = path

    @classmethod
    def set_attestation_report_path(cls, path):
        cls.ATTESTATION_REPORT_PATH = path

    @classmethod
    def set_gpu_attestation_certificates_path(cls, path):
        cls.GPU_ATTESTATION_CERTIFICATES_PATH = path


class BlackwellSettings(BaseSettings):
    signature_length = 96
    HashFunction = sha384
    MAX_CERT_CHAIN_LENGTH = 5
    HashFunctionNamespace = "{http://www.w3.org/2001/04/xmlenc#sha384}"
    GPU_ARCH_NAME = "BLACKWELL"
    GPU_ARCH = 10
    RIM_DIRECTORY_PATH = os.path.join(parent_dir, "samples")
    TEST_NO_GPU_DRIVER_RIM_PATH = ""
    DRIVER_RIM_PATH = ""
    TEST_NO_GPU_VBIOS_RIM_PATH = ""
    VBIOS_RIM_PATH = ""
    ATTESTATION_REPORT_PATH = ""
    GPU_ATTESTATION_CERTIFICATES_PATH = ""

    @classmethod
    def set_driver_rim_path(cls, path):
        cls.DRIVER_RIM_PATH = path

    @classmethod
    def set_vbios_rim_path(cls, path):
        cls.VBIOS_RIM_PATH = path

    @classmethod
    def set_attestation_report_path(cls, path):
        cls.ATTESTATION_REPORT_PATH = path

    @classmethod
    def set_gpu_attestation_certificates_path(cls, path):
        cls.GPU_ATTESTATION_CERTIFICATES_PATH = path


def dictionary_init():
    return None


GPU_ARCHITECTURE_MAP = defaultdict(dictionary_init)
GPU_ARCHITECTURE_MAP[9] = HopperSettings()
GPU_ARCHITECTURE_MAP[10] = BlackwellSettings()



#!/usr/bin/env python3
"""Offline evaluator for bulla.attestation-boundary/0.1-experimental."""

from __future__ import annotations

import argparse
import datetime as dt
import hashlib
import importlib.metadata
import importlib.util
import itertools
import json
from pathlib import Path
import re
import sys
from typing import Any, Mapping
import xml.etree.ElementTree as ET

sys.dont_write_bytecode = True


ROOT = Path(__file__).resolve().parent
VENDOR_LOCK = ROOT / "vendor-lock.json"
CONTEXT = ROOT / "context.json"
REPORT = ROOT / "generated/report.json"
PROFILE = "bulla.attestation-boundary/0.1-experimental"
REPORT_PROFILE = "bulla.attestation-boundary-report/0.1-experimental"
EXPECTED_SOURCE_LOCK_ROOT = "sha256:80dc5f6daa5a7445c18001825412281fab80cd8fbff17c44b18ae1fecbec1bb2"
EXPECTED_RUNTIME_DEPENDENCIES = {
    "PyNaCl": "1.6.2",
    "cbor2": "5.6.5",
    "cryptography": "43.0.1",
    "ecdsa": "0.18.0",
    "nvidia-ml-py": "12.550.52",
}
SUBCLAIMS = (
    "device_identity_and_state",
    "measured_workload",
    "model_commitment",
    "request_and_nonce_commitment",
    "response_commitment",
    "execution_relation",
)
STATES = {"ESTABLISHED", "NOT_ESTABLISHED", "UNCHECKABLE"}
RESULT_FIELDS = {
    "state", "reason_codes", "verification_depth", "accepted_roots",
    "accepted_policy_sha256", "evidence_sha256",
}


class AttestationBoundaryError(ValueError):
    pass


def canonical(value: Any) -> bytes:
    return (json.dumps(value, sort_keys=True, separators=(",", ":"), ensure_ascii=False) + "\n").encode()


def digest(raw: bytes) -> str:
    return "sha256:" + hashlib.sha256(raw).hexdigest()


def strict_json(path: Path) -> Any:
    def pairs(items: list[tuple[str, Any]]) -> dict[str, Any]:
        result: dict[str, Any] = {}
        for key, value in items:
            if key in result:
                raise AttestationBoundaryError(f"duplicate JSON member {key} in {path}")
            result[key] = value
        return result

    try:
        return json.loads(
            path.read_text(encoding="utf-8"),
            object_pairs_hook=pairs,
            parse_float=lambda _: (_ for _ in ()).throw(AttestationBoundaryError("floats are forbidden")),
            parse_constant=lambda _: (_ for _ in ()).throw(AttestationBoundaryError("non-finite numbers are forbidden")),
        )
    except (OSError, UnicodeDecodeError, json.JSONDecodeError) as exc:
        raise AttestationBoundaryError(f"invalid JSON at {path}: {exc}") from exc


def exact(value: Mapping[str, Any], fields: set[str], label: str) -> None:
    if set(value) != fields:
        raise AttestationBoundaryError(f"{label} fields changed")


def string_list(value: Any, label: str, *, nonempty: bool = False) -> None:
    if not isinstance(value, list) or (nonempty and not value):
        raise AttestationBoundaryError(f"{label} must be a{' nonempty' if nonempty else ''} string array")
    if any(not isinstance(item, str) or not item for item in value):
        raise AttestationBoundaryError(f"{label} contains an invalid string")


def sha256_value(value: Any, label: str) -> None:
    if not isinstance(value, str) or re.fullmatch(r"sha256:[0-9a-f]{64}", value) is None:
        raise AttestationBoundaryError(f"{label} is not a canonical SHA-256 digest")


def nonempty_string(value: Any, label: str) -> None:
    if not isinstance(value, str) or not value:
        raise AttestationBoundaryError(f"{label} must be a nonempty string")


def lowercase_hex(value: Any, label: str, *, length: int | None = None) -> None:
    pattern = rf"[0-9a-f]{{{length}}}" if length is not None else r"[0-9a-f]+"
    if not isinstance(value, str) or re.fullmatch(pattern, value) is None:
        raise AttestationBoundaryError(f"{label} must be canonical lowercase hexadecimal")


def safe_integer(value: Any, label: str) -> None:
    if type(value) is not int or not 0 <= value <= 9_007_199_254_740_991:
        raise AttestationBoundaryError(f"{label} must be a nonnegative safe integer")


def check_context(context: Mapping[str, Any]) -> None:
    exact(context, {"profile", "air", "nvidia", "live_h100_collection", "never_infer"}, "attestation context")
    if context["profile"] != "bulla.attestation-boundary-context/0.1-experimental":
        raise AttestationBoundaryError("attestation context profile changed")
    air = context["air"]
    nvidia = context["nvidia"]
    if not isinstance(air, Mapping) or not isinstance(nvidia, Mapping):
        raise AttestationBoundaryError("attestation contexts must be objects")
    exact(air, {
        "accepted_vector", "accepted_public_key_hex", "expected_model_hash_hex", "expected_nonce_hex",
        "expected_platform", "frozen_now_epoch",
    }, "AIR context")
    exact(nvidia, {
        "accepted_device_root_der_sha256", "evaluation_time", "expected_report_format", "expected_driver_version",
        "expected_measurement_count", "expected_sample_nonce_hex", "expected_vbios_version", "recorded_state_semantics",
    }, "NVIDIA context")
    accepted_vector = air["accepted_vector"]
    if (
        not isinstance(accepted_vector, str)
        or re.fullmatch(r"vendor/air/vectors/valid/[a-z0-9.-]+\.json", accepted_vector) is None
        or ".." in Path(accepted_vector).parts
    ):
        raise AttestationBoundaryError("AIR accepted vector path is not canonical")
    lowercase_hex(air["accepted_public_key_hex"], "AIR accepted public key", length=64)
    lowercase_hex(air["expected_model_hash_hex"], "AIR expected model hash", length=64)
    lowercase_hex(air["expected_nonce_hex"], "AIR expected nonce")
    nonempty_string(air["expected_platform"], "AIR expected platform")
    safe_integer(air["frozen_now_epoch"], "AIR frozen evaluation time")
    sha256_value(nvidia["accepted_device_root_der_sha256"], "NVIDIA accepted device root")
    evaluation_time = nvidia["evaluation_time"]
    if not isinstance(evaluation_time, str) or re.fullmatch(
        r"[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}Z", evaluation_time
    ) is None:
        raise AttestationBoundaryError("NVIDIA evaluation time must use canonical UTC Z form")
    try:
        dt.datetime.fromisoformat(evaluation_time.replace("Z", "+00:00"))
    except ValueError as exc:
        raise AttestationBoundaryError("NVIDIA evaluation time is invalid") from exc
    nonempty_string(nvidia["expected_driver_version"], "NVIDIA expected driver version")
    safe_integer(nvidia["expected_measurement_count"], "NVIDIA expected measurement count")
    lowercase_hex(nvidia["expected_sample_nonce_hex"], "NVIDIA expected sample nonce", length=64)
    nonempty_string(nvidia["expected_vbios_version"], "NVIDIA expected VBIOS version")
    if nvidia["recorded_state_semantics"] != "SIGNED_REPORT_CONTENTS_NOT_RIM_CONFORMITY":
        raise AttestationBoundaryError("NVIDIA recorded-state boundary changed")
    if nvidia["expected_report_format"] != "HOPPER_ATTESTATION_REPORT":
        raise AttestationBoundaryError("NVIDIA accepted report parser profile changed")
    if context["live_h100_collection"] != "UNCHECKABLE":
        raise AttestationBoundaryError("live H100 collection boundary changed")
    string_list(context["never_infer"], "context never_infer", nonempty=True)


def check_vendor_lock() -> Mapping[str, Any]:
    lock = strict_json(VENDOR_LOCK)
    if not isinstance(lock, Mapping):
        raise AttestationBoundaryError("vendor lock must be an object")
    exact(lock, {"profile", "sources", "runtime_dependencies", "members", "lock_root"}, "vendor lock")
    if lock["profile"] != "bulla.attestation-boundary-source-lock/1":
        raise AttestationBoundaryError("vendor lock profile changed")
    if lock["runtime_dependencies"] != EXPECTED_RUNTIME_DEPENDENCIES:
        raise AttestationBoundaryError("accepted verifier dependency versions changed")
    for distribution, expected_version in EXPECTED_RUNTIME_DEPENDENCIES.items():
        try:
            actual_version = importlib.metadata.version(distribution)
        except importlib.metadata.PackageNotFoundError as exc:
            raise AttestationBoundaryError(f"required verifier dependency is absent: {distribution}") from exc
        if actual_version != expected_version:
            raise AttestationBoundaryError(
                f"verifier dependency changed: {distribution} {actual_version} != {expected_version}"
            )
    body = {key: value for key, value in lock.items() if key != "lock_root"}
    if lock["lock_root"] != digest(canonical(body)):
        raise AttestationBoundaryError("vendor lock root changed")
    if lock["lock_root"] != EXPECTED_SOURCE_LOCK_ROOT:
        raise AttestationBoundaryError("vendor lock does not match the receiver-accepted source closure")
    sources = lock["sources"]
    if sources != {
        "air": {"commit": "e3bacb3076338db4f9644e469410a6ffea11c303", "release": "v1.0.1"},
        "nvtrust": {"commit": "0c5d627313037c1e577d05a232e79394a41b2c21"},
    }:
        raise AttestationBoundaryError("accepted upstream source commits changed")
    members = lock["members"]
    if not isinstance(members, list) or not members:
        raise AttestationBoundaryError("vendor lock members must be a nonempty array")
    declared: list[str] = []
    for index, member in enumerate(members):
        if not isinstance(member, Mapping):
            raise AttestationBoundaryError(f"vendor member {index} must be an object")
        exact(member, {
            "path", "source_repository", "source_commit", "source_path", "license", "byte_length", "sha256"
        }, f"vendor member {index}")
        path = member["path"]
        if not isinstance(path, str) or not path.startswith("vendor/") or ".." in Path(path).parts:
            raise AttestationBoundaryError("vendor path is not canonical")
        if declared and path <= declared[-1]:
            raise AttestationBoundaryError("vendor paths are not strict ASCII order")
        declared.append(path)
        if path.startswith("vendor/air/"):
            expected_source = {
                "source_repository": "https://github.com/cyntrisec/air-v1",
                "source_commit": sources["air"]["commit"],
                "source_path": path.removeprefix("vendor/air/"),
                "license": "CC-BY-4.0",
            }
        elif path.startswith("vendor/nvtrust/"):
            expected_source = {
                "source_repository": "https://github.com/NVIDIA/nvtrust",
                "source_commit": sources["nvtrust"]["commit"],
                "source_path": path.removeprefix("vendor/nvtrust/"),
                "license": "BSD-3-Clause-or-bundled-notices" if "/local_gpu_verifier/" in path else "Apache-2.0",
            }
        else:
            raise AttestationBoundaryError("vendor path has no accepted source jurisdiction")
        if any(member[key] != value for key, value in expected_source.items()):
            raise AttestationBoundaryError(f"vendor source attribution changed: {path}")
        target = ROOT / path
        if target.is_symlink() or not target.is_file():
            raise AttestationBoundaryError(f"vendor member is missing or non-regular: {path}")
        raw = target.read_bytes()
        if type(member["byte_length"]) is not int or member["byte_length"] != len(raw):
            raise AttestationBoundaryError(f"vendor member length changed: {path}")
        if member["sha256"] != digest(raw):
            raise AttestationBoundaryError(f"vendor member digest changed: {path}")
    actual = sorted(
        path.relative_to(ROOT).as_posix()
        for path in (ROOT / "vendor").rglob("*")
        if path.is_file() or path.is_symlink()
    )
    if actual != declared:
        raise AttestationBoundaryError("vendor member closure changed")
    return lock


def result(
    state: str,
    reason_codes: list[str],
    verification_depth: list[str],
    accepted_roots: list[str],
    accepted_policy_sha256: str,
    evidence_sha256: list[str],
) -> dict[str, Any]:
    if state not in STATES or not reason_codes:
        raise AttestationBoundaryError("subclaim result is not closed")
    return {
        "state": state,
        "reason_codes": reason_codes,
        "verification_depth": verification_depth,
        "accepted_roots": accepted_roots,
        "accepted_policy_sha256": accepted_policy_sha256,
        "evidence_sha256": evidence_sha256,
    }


def load_air_module():
    path = ROOT / "vendor/air/scripts/interop_test.py"
    spec = importlib.util.spec_from_file_location("attestation_boundary_air", path)
    if spec is None or spec.loader is None:
        raise AttestationBoundaryError("cannot load the pinned AIR verifier")
    module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module
    spec.loader.exec_module(module)
    return module


def evaluate_air(context: Mapping[str, Any], policy_digest: str) -> tuple[dict[str, Any], dict[str, Any]]:
    air = load_air_module()
    vector_root = ROOT / "vendor/air/vectors"
    vector_results: list[dict[str, Any]] = []
    for path in sorted((vector_root / "valid").glob("*.json")) + sorted((vector_root / "invalid").glob("*.json")):
        vector = strict_json(path)
        verified = air.verify_vector(dict(vector), context["frozen_now_epoch"])
        if "expected_failure" in vector:
            failure = verified.failure
            expected = vector["expected_failure"]
            passed = (
                not verified.ok
                and failure is not None
                and failure.layer == expected["layer"]
                and failure.check == expected["check"]
                and failure.code == expected["code"]
            )
        else:
            passed = verified.ok
        if not passed:
            raise AttestationBoundaryError(f"pinned AIR vector outcome changed: {path.name}")
        vector_results.append({"name": vector["name"], "expected_outcome_reproduced": True})

    accepted_path = ROOT / context["accepted_vector"]
    accepted = strict_json(accepted_path)
    if accepted["public_key_hex"] != context["accepted_public_key_hex"]:
        raise AttestationBoundaryError("AIR accepted key changed")
    policy_vector = dict(accepted)
    policy_vector["verify_policy"] = {
        "expected_model_hash_hex": context["expected_model_hash_hex"],
        "expected_nonce_hex": context["expected_nonce_hex"],
        "expected_platform": context["expected_platform"],
        "require_nonce": True,
    }
    verification = air.verify_vector(policy_vector, context["frozen_now_epoch"])
    if not verification.ok:
        raise AttestationBoundaryError("AIR accepted vector failed the receiver policy")
    protected, _, payload, signature = air.parse_receipt_cose(bytes.fromhex(accepted["receipt_hex"]))
    claims = air.decode_and_validate_claims(payload)
    air.verify_sig_structure(protected, payload, signature, bytes.fromhex(context["accepted_public_key_hex"]))
    evidence = [digest(accepted_path.read_bytes())]
    root = digest(bytes.fromhex(context["accepted_public_key_hex"]))
    signed_depth = ["COSE_SIGN1", "ED25519", "AIR_V1_CLAIMS", "RECEIVER_PINNED_POLICY"]
    signed = lambda reasons: result("ESTABLISHED", reasons, signed_depth, [root], policy_digest, evidence)
    not_execution = lambda reasons: result("NOT_ESTABLISHED", reasons, signed_depth, [root], policy_digest, evidence)
    subclaims = {
        "device_identity_and_state": not_execution([
            "SIGNED_PLATFORM_AND_MEASUREMENT_CLAIMS_BOUND",
            "SIGNER_KEY_NOT_BOUND_TO_DEVICE_IDENTITY",
        ]),
        "measured_workload": not_execution(["PLATFORM_MEASUREMENTS_DO_NOT_BIND_WORKLOAD_EXECUTION"]),
        "model_commitment": signed(["SIGNED_MODEL_COMMITMENT_BOUND"]),
        "request_and_nonce_commitment": signed(["SIGNED_REQUEST_AND_NONCE_COMMITMENTS_BOUND"]),
        "response_commitment": signed(["SIGNED_RESPONSE_COMMITMENT_BOUND"]),
        "execution_relation": not_execution(["NO_VERIFIED_INFERENCE_RELATION"]),
    }
    details = {
        "accepted_vector": accepted_path.relative_to(ROOT).as_posix(),
        "accepted_key_sha256": root,
        "model_hash_hex": bytes(claims[air.AIR_MODEL_HASH]).hex(),
        "request_hash_hex": bytes(claims[air.AIR_REQUEST_HASH]).hex(),
        "response_hash_hex": bytes(claims[air.AIR_RESPONSE_HASH]).hex(),
        "nonce_hex": bytes(claims[air.EAT_NONCE]).hex(),
        "platform_claim": claims[air.AIR_ENCLAVE_MEASUREMENTS]["measurement_type"],
        "vectors": vector_results,
    }
    return subclaims, details


def verify_certificate_signature(certificate: Any, issuer: Any) -> None:
    from cryptography.hazmat.primitives.asymmetric import ec
    try:
        issuer.public_key().verify(
            certificate.signature,
            certificate.tbs_certificate_bytes,
            ec.ECDSA(certificate.signature_hash_algorithm),
        )
    except Exception as exc:
        raise AttestationBoundaryError("NVIDIA certificate signature verification failed") from exc


def rim_measurements(path: Path, runtime: list[str]) -> dict[str, Any]:
    try:
        root = ET.fromstring(path.read_bytes())
    except ET.ParseError as exc:
        raise AttestationBoundaryError(f"pinned RIM XML is invalid: {path.name}") from exc
    metadata = next((element for element in root.iter() if element.tag.endswith("Meta")), None)
    if metadata is None:
        raise AttestationBoundaryError(f"pinned RIM has no metadata: {path.name}")
    active: list[int] = []
    mismatched: list[int] = []
    namespace = "{http://www.w3.org/2001/04/xmlenc#sha384}"
    for element in root.iter():
        if element.attrib.get("type") != "Measurement" or element.attrib.get("active") != "True":
            continue
        index = int(element.attrib["index"])
        alternatives = int(element.attrib["alternatives"])
        accepted = [element.attrib[f"{namespace}Hash{offset}"] for offset in range(alternatives)]
        active.append(index)
        if index >= len(runtime) or runtime[index] not in accepted:
            mismatched.append(index)
    return {
        "sha256": digest(path.read_bytes()),
        "declared_version": metadata.attrib.get("colloquialVersion"),
        "active_measurements": active,
        "mismatched_measurements": mismatched,
    }


def evaluate_nvidia(context: Mapping[str, Any], policy_digest: str) -> tuple[dict[str, Any], dict[str, Any]]:
    from cryptography import x509
    from cryptography.hazmat.primitives import serialization

    local = ROOT / "vendor/nvtrust/guest_tools/gpu_verifiers/local_gpu_verifier"
    source = local / "src"
    sys.path.insert(0, str(source))
    try:
        from verifier.attestation import AttestationReport
        from verifier.config import HopperSettings
        from verifier.utils import format_vbios_version
    finally:
        sys.path.pop(0)

    sample = local / "src/verifier/samples"
    certs = local / "src/verifier/certs"
    report_path = sample / "hopperAttestationReport.txt"
    chain_path = sample / "hopperCertChain.txt"
    root_path = certs / "verifier_device_root.pem"
    report_raw = bytes.fromhex(report_path.read_text(encoding="ascii"))
    report = AttestationReport(report_raw, HopperSettings())
    blocks = re.findall(
        b"-----BEGIN CERTIFICATE-----.*?-----END CERTIFICATE-----",
        chain_path.read_bytes(),
        re.DOTALL,
    )
    chain = [x509.load_pem_x509_certificate(block) for block in blocks]
    accepted_root = x509.load_pem_x509_certificate(root_path.read_bytes())
    if len(chain) != 5 or chain[-1].public_bytes(serialization.Encoding.DER) != accepted_root.public_bytes(serialization.Encoding.DER):
        raise AttestationBoundaryError("NVIDIA device chain does not terminate at the accepted root")
    root_digest = digest(accepted_root.public_bytes(serialization.Encoding.DER))
    if root_digest != context["accepted_device_root_der_sha256"]:
        raise AttestationBoundaryError("NVIDIA accepted device root changed")
    evaluation_time = dt.datetime.fromisoformat(context["evaluation_time"].replace("Z", "+00:00"))
    for index, certificate in enumerate(chain):
        issuer = chain[index + 1] if index + 1 < len(chain) else certificate
        if certificate.issuer != issuer.subject:
            raise AttestationBoundaryError("NVIDIA certificate issuer chain changed")
        if not (certificate.not_valid_before_utc <= evaluation_time <= certificate.not_valid_after_utc):
            raise AttestationBoundaryError("NVIDIA device certificate is outside the accepted evaluation time")
        verify_certificate_signature(certificate, issuer)
    if not report.verify_signature(chain[0], HopperSettings.signature_length, hashlib.sha384):
        raise AttestationBoundaryError("NVIDIA sample report signature failed")
    request_nonce = report.get_request_message().get_nonce().hex()
    if request_nonce != context["expected_sample_nonce_hex"]:
        raise AttestationBoundaryError("NVIDIA sample nonce changed")
    opaque = report.get_response_message().get_opaque_data()
    driver = opaque.get_data("OPAQUE_FIELD_ID_DRIVER_VERSION").decode().rstrip("\0")
    vbios = format_vbios_version(opaque.get_data("OPAQUE_FIELD_ID_VBIOS_VERSION"))
    if driver != context["expected_driver_version"] or vbios != context["expected_vbios_version"]:
        raise AttestationBoundaryError("NVIDIA recorded driver or VBIOS version changed")
    runtime = report.get_measurements()
    if len(runtime) != context["expected_measurement_count"]:
        raise AttestationBoundaryError("NVIDIA measurement count changed")
    driver_rim = rim_measurements(sample / "Driver_RIM_test_no_gpu.swidtag", runtime)
    vbios_rim = rim_measurements(sample / "1010_0200_882_96005E0001_test_no_gpu.swidtag", runtime)
    evidence = [
        digest(report_path.read_bytes()),
        digest(chain_path.read_bytes()),
        driver_rim["sha256"],
        vbios_rim["sha256"],
    ]
    signed_depth = ["PINNED_DEVICE_CERTIFICATE_CHAIN", "SPDM_REPORT_SIGNATURE", "RECORDED_SAMPLE_NONCE", "MEASUREMENT_RECORD_PARSE"]
    subclaims = {
        "device_identity_and_state": result(
            "ESTABLISHED",
            ["PINNED_DEVICE_CHAIN_VALID", "RECORDED_REPORT_SIGNATURE_VALID", "SIGNED_RECORDED_STATE_BOUND"],
            signed_depth,
            [root_digest],
            policy_digest,
            evidence,
        ),
        "measured_workload": result(
            "NOT_ESTABLISHED",
            ["RIM_POLICY_MISMATCH", "FIRMWARE_MEASUREMENTS_DO_NOT_BIND_WORKLOAD_EXECUTION"],
            signed_depth,
            [root_digest],
            policy_digest,
            evidence,
        ),
        "model_commitment": result("NOT_ESTABLISHED", ["MODEL_COMMITMENT_ABSENT"], signed_depth, [root_digest], policy_digest, evidence),
        "request_and_nonce_commitment": result("NOT_ESTABLISHED", ["REQUEST_COMMITMENT_ABSENT", "SAMPLE_NONCE_NOT_LIVE_CHALLENGE"], signed_depth, [root_digest], policy_digest, evidence),
        "response_commitment": result("NOT_ESTABLISHED", ["RESPONSE_COMMITMENT_ABSENT"], signed_depth, [root_digest], policy_digest, evidence),
        "execution_relation": result("NOT_ESTABLISHED", ["MODEL_REQUEST_RESPONSE_RELATION_ABSENT"], signed_depth, [root_digest], policy_digest, evidence),
    }
    details = {
        "report_format": context["expected_report_format"],
        "report_format_jurisdiction": "PINNED_PARSER_PROFILE_NOT_SIGNED_REPORT_FIELD",
        "device_root_der_sha256": root_digest,
        "report_sha256": digest(report_path.read_bytes()),
        "certificate_chain_sha256": digest(chain_path.read_bytes()),
        "sample_nonce_hex": request_nonce,
        "driver_version": driver,
        "vbios_version": vbios,
        "measurement_count": len(runtime),
        "measurements_sha256": digest(canonical(runtime)),
        "driver_rim": driver_rim,
        "vbios_rim": vbios_rim,
        "rim_signature_evaluation": "NOT_PERFORMED",
    }
    return subclaims, details


def minimal_sets(sets: list[frozenset[str]]) -> list[list[str]]:
    unique = sorted(set(sets), key=lambda item: (len(item), tuple(sorted(item))))
    minimal = [item for item in unique if not any(other < item for other in unique)]
    return [sorted(item) for item in minimal]


def acceptance_cuts(nodes: list[Mapping[str, Any]], edges: list[Mapping[str, str]]) -> list[dict[str, Any]]:
    by_id = {node["id"]: node for node in nodes}
    dependencies: dict[str, list[str]] = {}
    for edge in edges:
        dependencies.setdefault(edge["to"], []).append(edge["from"])

    def paths(node_id: str) -> list[frozenset[str]]:
        node = by_id[node_id]
        if node["kind"] == "CONTROL":
            return [frozenset({node_id})]
        if node["state"] != "ESTABLISHED":
            return []
        child_sets = [paths(child) for child in dependencies.get(node_id, [])]
        if not child_sets or any(not choices for choices in child_sets):
            return []
        return [frozenset().union(*choice) for choice in itertools.product(*child_sets)]

    results: list[dict[str, Any]] = []
    for node in nodes:
        if node["kind"] == "CONCLUSION":
            results.append({"conclusion": node["id"], "state": node["state"], "cuts": minimal_sets(paths(node["id"]))})
    return results


def dependency_graph(air: Mapping[str, Any], nvidia: Mapping[str, Any]) -> tuple[dict[str, Any], list[dict[str, Any]]]:
    controls = {
        "air.accepted_key", "air.accepted_policy", "air.vector_bytes", "air.verifier_source",
        "air.runtime_dependency_versions", "nvidia.certificate_chain", "nvidia.device_root",
        "nvidia.accepted_policy", "nvidia.report_bytes", "nvidia.report_parser_source", "nvidia.runtime_dependency_versions",
    }
    nodes: list[dict[str, Any]] = [{"id": item, "kind": "CONTROL"} for item in sorted(controls)]
    edges: list[dict[str, str]] = []
    dependencies = {
        "air": [
            "air.accepted_key", "air.accepted_policy", "air.vector_bytes", "air.verifier_source",
            "air.runtime_dependency_versions",
        ],
        "nvidia": [
            "nvidia.certificate_chain", "nvidia.device_root", "nvidia.accepted_policy", "nvidia.report_bytes",
            "nvidia.report_parser_source", "nvidia.runtime_dependency_versions",
        ],
    }
    for artifact, subclaims in (("air", air), ("nvidia", nvidia)):
        for name in SUBCLAIMS:
            conclusion = f"{artifact}.{name}"
            nodes.append({"id": conclusion, "kind": "CONCLUSION", "state": subclaims[name]["state"]})
            for control in dependencies[artifact]:
                edges.append({"from": control, "to": conclusion})
    nodes.sort(key=lambda item: item["id"])
    edges.sort(key=lambda item: (item["to"], item["from"]))
    graph = {"logic": "ALL_DECLARED_CONTROLS_REQUIRED", "nodes": nodes, "edges": edges}
    return graph, acceptance_cuts(nodes, edges)


def build_report() -> dict[str, Any]:
    lock = check_vendor_lock()
    context = strict_json(CONTEXT)
    if not isinstance(context, Mapping):
        raise AttestationBoundaryError("attestation context must be an object")
    check_context(context)
    policy_digest = digest(canonical(context))
    air, air_details = evaluate_air(context["air"], policy_digest)
    nvidia, nvidia_details = evaluate_nvidia(context["nvidia"], policy_digest)
    for artifact in (air, nvidia):
        if set(artifact) != set(SUBCLAIMS):
            raise AttestationBoundaryError("attestation subclaim closure changed")
    graph, cuts = dependency_graph(air, nvidia)
    report: dict[str, Any] = {
        "profile": REPORT_PROFILE,
        "source_profile": PROFILE,
        "source_lock_root": lock["lock_root"],
        "accepted_policy_sha256": policy_digest,
        "upstream_commits": {
            "air": lock["sources"]["air"]["commit"],
            "nvtrust": lock["sources"]["nvtrust"]["commit"],
        },
        "artifacts": {
            "air": {"subclaims": air, "details": air_details},
            "nvidia": {"subclaims": nvidia, "details": nvidia_details},
        },
        "combined_execution_relation": result(
            "NOT_ESTABLISHED",
            ["NO_AUTHENTICATED_SHARED_EVENT_BINDING", "NO_VERIFIED_INFERENCE_RELATION"],
            ["SEPARATE_AIR_AND_NVIDIA_VERIFICATION"],
            [],
            policy_digest,
            sorted(set(air["execution_relation"]["evidence_sha256"] + nvidia["execution_relation"]["evidence_sha256"])),
        ),
        "live_h100_collection": result(
            "UNCHECKABLE",
            ["LIVE_H100_EVIDENCE_NOT_SUPPLIED"],
            [],
            [],
            policy_digest,
            [],
        ),
        "control_dependency_graph": graph,
        "acceptance_cuts": cuts,
        "never_infer": context["never_infer"],
    }
    report["report_sha256"] = digest(canonical(report))
    return report


def check_report(report: Mapping[str, Any]) -> None:
    exact(report, {
        "profile", "source_profile", "source_lock_root", "accepted_policy_sha256", "upstream_commits",
        "artifacts", "combined_execution_relation", "live_h100_collection", "control_dependency_graph",
        "acceptance_cuts", "never_infer", "report_sha256",
    }, "attestation boundary report")
    if report["profile"] != REPORT_PROFILE or report["source_profile"] != PROFILE:
        raise AttestationBoundaryError("attestation boundary report profile changed")
    if report["source_lock_root"] != EXPECTED_SOURCE_LOCK_ROOT:
        raise AttestationBoundaryError("attestation boundary source lock changed")
    if report["upstream_commits"] != {
        "air": "e3bacb3076338db4f9644e469410a6ffea11c303",
        "nvtrust": "0c5d627313037c1e577d05a232e79394a41b2c21",
    }:
        raise AttestationBoundaryError("attestation boundary upstream commits changed")
    sha256_value(report["accepted_policy_sha256"], "accepted policy")
    sha256_value(report["report_sha256"], "report")
    body = {key: value for key, value in report.items() if key != "report_sha256"}
    if report["report_sha256"] != digest(canonical(body)):
        raise AttestationBoundaryError("attestation boundary report digest changed")
    artifacts = report["artifacts"]
    if not isinstance(artifacts, Mapping) or set(artifacts) != {"air", "nvidia"}:
        raise AttestationBoundaryError("attestation artifact set changed")
    for artifact_name in ("air", "nvidia"):
        artifact = artifacts[artifact_name]
        if not isinstance(artifact, Mapping):
            raise AttestationBoundaryError(f"{artifact_name} artifact must be an object")
        exact(artifact, {"subclaims", "details"}, f"{artifact_name} artifact")
        if not isinstance(artifact["subclaims"], Mapping) or set(artifact["subclaims"]) != set(SUBCLAIMS):
            raise AttestationBoundaryError(f"{artifact_name} subclaim closure changed")
        for name, claim in artifact["subclaims"].items():
            if not isinstance(claim, Mapping):
                raise AttestationBoundaryError(f"{artifact_name}.{name} must be an object")
            exact(claim, RESULT_FIELDS, f"{artifact_name}.{name}")
            if claim["state"] not in STATES:
                raise AttestationBoundaryError(f"{artifact_name}.{name} state changed")
            string_list(claim["reason_codes"], f"{artifact_name}.{name}.reason_codes", nonempty=True)
            for field in ("verification_depth", "accepted_roots", "evidence_sha256"):
                string_list(claim[field], f"{artifact_name}.{name}.{field}")
            for field in ("accepted_roots", "evidence_sha256"):
                for value in claim[field]:
                    sha256_value(value, f"{artifact_name}.{name}.{field}")
            if claim["accepted_policy_sha256"] != report["accepted_policy_sha256"]:
                raise AttestationBoundaryError(f"{artifact_name}.{name} policy binding changed")
        details = artifact["details"]
        if not isinstance(details, Mapping):
            raise AttestationBoundaryError(f"{artifact_name} details must be an object")
        if artifact_name == "air":
            exact(details, {
                "accepted_vector", "accepted_key_sha256", "model_hash_hex", "request_hash_hex",
                "response_hash_hex", "nonce_hex", "platform_claim", "vectors",
            }, "AIR details")
            if not isinstance(details["vectors"], list) or len(details["vectors"]) != 10:
                raise AttestationBoundaryError("AIR vector result closure changed")
            for vector in details["vectors"]:
                if not isinstance(vector, Mapping):
                    raise AttestationBoundaryError("AIR vector result must be an object")
                exact(vector, {"name", "expected_outcome_reproduced"}, "AIR vector result")
                if not isinstance(vector["name"], str) or vector["expected_outcome_reproduced"] is not True:
                    raise AttestationBoundaryError("AIR vector result changed")
            sha256_value(details["accepted_key_sha256"], "AIR accepted key")
            nonempty_string(details["accepted_vector"], "AIR accepted vector")
            for field in ("model_hash_hex", "request_hash_hex", "response_hash_hex", "nonce_hex"):
                lowercase_hex(details[field], f"AIR {field}")
            nonempty_string(details["platform_claim"], "AIR platform claim")
        else:
            exact(details, {
                "report_format", "report_format_jurisdiction", "device_root_der_sha256", "report_sha256", "certificate_chain_sha256",
                "sample_nonce_hex", "driver_version", "vbios_version", "measurement_count",
                "measurements_sha256", "driver_rim", "vbios_rim", "rim_signature_evaluation",
            }, "NVIDIA details")
            if type(details["measurement_count"]) is not int:
                raise AttestationBoundaryError("NVIDIA measurement count must be an exact integer")
            safe_integer(details["measurement_count"], "NVIDIA measurement count")
            if details["report_format"] != "HOPPER_ATTESTATION_REPORT" or details["report_format_jurisdiction"] != "PINNED_PARSER_PROFILE_NOT_SIGNED_REPORT_FIELD":
                raise AttestationBoundaryError("NVIDIA report-format jurisdiction changed")
            for field in ("device_root_der_sha256", "report_sha256", "certificate_chain_sha256", "measurements_sha256"):
                sha256_value(details[field], f"NVIDIA {field}")
            lowercase_hex(details["sample_nonce_hex"], "NVIDIA sample nonce")
            nonempty_string(details["driver_version"], "NVIDIA driver version")
            nonempty_string(details["vbios_version"], "NVIDIA VBIOS version")
            if details["rim_signature_evaluation"] != "NOT_PERFORMED":
                raise AttestationBoundaryError("NVIDIA RIM signature boundary changed")
            for label in ("driver_rim", "vbios_rim"):
                rim = details[label]
                if not isinstance(rim, Mapping):
                    raise AttestationBoundaryError(f"{label} must be an object")
                exact(rim, {"sha256", "declared_version", "active_measurements", "mismatched_measurements"}, label)
                sha256_value(rim["sha256"], f"{label}.sha256")
                if rim["declared_version"] is not None:
                    nonempty_string(rim["declared_version"], f"{label}.declared_version")
                for field in ("active_measurements", "mismatched_measurements"):
                    values = rim[field]
                    if not isinstance(values, list):
                        raise AttestationBoundaryError(f"{label}.{field} must be an exact integer array")
                    for value in values:
                        safe_integer(value, f"{label}.{field}")
    for label in ("combined_execution_relation", "live_h100_collection"):
        claim = report[label]
        if not isinstance(claim, Mapping):
            raise AttestationBoundaryError(f"{label} must be an object")
        exact(claim, RESULT_FIELDS, label)
        if claim["state"] not in STATES:
            raise AttestationBoundaryError(f"{label} state changed")
        string_list(claim["reason_codes"], f"{label}.reason_codes", nonempty=True)
        for field in ("verification_depth", "accepted_roots", "evidence_sha256"):
            string_list(claim[field], f"{label}.{field}")
        for field in ("accepted_roots", "evidence_sha256"):
            for value in claim[field]:
                sha256_value(value, f"{label}.{field}")
        if claim["accepted_policy_sha256"] != report["accepted_policy_sha256"]:
            raise AttestationBoundaryError(f"{label} policy binding changed")
    graph = report["control_dependency_graph"]
    if not isinstance(graph, Mapping):
        raise AttestationBoundaryError("control dependency graph must be an object")
    exact(graph, {"logic", "nodes", "edges"}, "control dependency graph")
    if graph["logic"] != "ALL_DECLARED_CONTROLS_REQUIRED":
        raise AttestationBoundaryError("control dependency logic changed")
    if not isinstance(graph["nodes"], list) or not isinstance(graph["edges"], list):
        raise AttestationBoundaryError("control dependency members must be arrays")
    for node in graph["nodes"]:
        if not isinstance(node, Mapping) or set(node) not in ({"id", "kind"}, {"id", "kind", "state"}):
            raise AttestationBoundaryError("control dependency node fields changed")
        if node["kind"] not in {"CONTROL", "CONCLUSION"}:
            raise AttestationBoundaryError("control dependency node kind changed")
        if node["kind"] == "CONCLUSION" and node.get("state") not in STATES:
            raise AttestationBoundaryError("control dependency conclusion state changed")
    for edge in graph["edges"]:
        if not isinstance(edge, Mapping):
            raise AttestationBoundaryError("control dependency edge must be an object")
        exact(edge, {"from", "to"}, "control dependency edge")
    if not isinstance(report["acceptance_cuts"], list):
        raise AttestationBoundaryError("acceptance cuts must be an array")
    for cut in report["acceptance_cuts"]:
        if not isinstance(cut, Mapping):
            raise AttestationBoundaryError("acceptance cut must be an object")
        exact(cut, {"conclusion", "state", "cuts"}, "acceptance cut")
        if cut["state"] not in STATES or not isinstance(cut["cuts"], list):
            raise AttestationBoundaryError("acceptance cut state or members changed")
        for controls in cut["cuts"]:
            string_list(controls, "acceptance cut controls")
    string_list(report["never_infer"], "report never_infer", nonempty=True)
    expected_graph, expected_cuts = dependency_graph(
        artifacts["air"]["subclaims"], artifacts["nvidia"]["subclaims"]
    )
    if graph != expected_graph or report["acceptance_cuts"] != expected_cuts:
        raise AttestationBoundaryError("control dependency graph or acceptance cuts changed")
    context = strict_json(CONTEXT)
    if not isinstance(context, Mapping):
        raise AttestationBoundaryError("attestation context must be an object")
    check_context(context)
    if report["accepted_policy_sha256"] != digest(canonical(context)) or report["never_infer"] != context["never_infer"]:
        raise AttestationBoundaryError("report context binding changed")


def main(argv: list[str] | None = None) -> int:
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("--report", type=Path, default=REPORT)
    arguments = parser.parse_args(argv)
    try:
        expected = build_report()
        actual = strict_json(arguments.report)
        if actual != expected:
            raise AttestationBoundaryError("generated attestation boundary report drifted")
        check_report(actual)
    except (AttestationBoundaryError, OSError, ImportError) as exc:
        print(f"attestation boundary rejected: {exc}", file=sys.stderr)
        return 1
    print(f"attestation boundary: {actual['report_sha256']} · PASS")
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



"""``jsonschema+quantum/1`` — the one executable predicate form, single-sourced.

A closed, decidable, stdlib-only constraint language over a flat object (the act's
declared ``subject``): a JSON-schema keyword subset plus an integer unit/quantum
declaration. Deliberately **not** a general language — every keyword below is
decidable by a verifier with no dependencies, so any party recomputes conformance
from the receipt alone.

This module is a **leaf** (it imports only :mod:`bulla._canonical`), so it can be
shared by every layer that needs the same predicate without a circular import:

- **conventions** (`action_receipt.py`) — rules coined at a seam;
- **bounds scope** (`envelope.py` + `action_receipt.py`) — the capability an act is
  permitted to have, so ``bounds_conformance`` can recompute *did the act obey its
  scope* rather than merely *does the chain convey a matching digest*.

Callers wrap :class:`ExecutableFormError` in their own error type
(``ActionReceiptError`` / ``EnvelopeError``) so existing exception contracts hold.
"""

from __future__ import annotations

import hashlib
import math
import re
from typing import Any

from bulla._canonical import canonical_json

#: The one executable-definition form. A definition naming any other form is malformed.
EXECUTABLE_FORM = "jsonschema+quantum/1"

#: The closed keyword vocabulary. Anything outside these sets is malformed — fail
#: closed, never guess (an unrecognized keyword could silently weaken a predicate).
_SCHEMA_TOP_KEYS = frozenset({"type", "properties", "required", "additionalProperties"})
_SCHEMA_PROP_KEYS = frozenset({"type", "enum", "const", "minimum", "maximum", "pattern"})
_SCHEMA_PROP_TYPES = frozenset({"string", "integer", "number", "boolean"})
_PORTABLE_PROP_TYPES = frozenset({"string", "integer", "boolean"})
_MAX_SAFE_INTEGER = 2**53 - 1


class ExecutableFormError(ValueError):
    """A definition is not a well-formed ``jsonschema+quantum/1`` predicate."""


def _sha(b: bytes) -> str:
    return f"sha256:{hashlib.sha256(b).hexdigest()}"


def definition_hash(definition: Any) -> str:
    """The pin: ``sha256:`` over ``canonical_json`` for a structured (executable)
    definition, over raw UTF-8 bytes for an opaque string. This polymorphism is what
    lets a scope be *either* prose (hashed as its UTF-8 bytes, byte-identical to the
    historical behaviour) *or* a structured predicate (hashed canonically) without a
    version fork in the hash."""
    if isinstance(definition, str):
        return _sha(definition.encode("utf-8"))
    return _sha(canonical_json(definition).encode("utf-8"))


def validate_executable_definition(defn: Any) -> None:
    """Well-formedness of a ``jsonschema+quantum/1`` definition — the closed
    vocabulary is validated at CONSTRUCTION so a malformed predicate can never ride
    inside a hashed receipt. Raises :class:`ExecutableFormError` on the first
    violation."""
    if not isinstance(defn, dict):
        raise ExecutableFormError("executable definition must be an object")
    if defn.get("form") != EXECUTABLE_FORM:
        raise ExecutableFormError(
            f"executable definition.form must be {EXECUTABLE_FORM!r} "
            "(the one declared form — not a general language)"
        )
    extra = set(defn) - {"form", "schema", "quantum"}
    if extra:
        raise ExecutableFormError(f"unknown executable-definition keys {sorted(extra)} — fail closed")
    schema = defn.get("schema")
    if not isinstance(schema, dict):
        raise ExecutableFormError("executable definition needs a 'schema' object")
    unknown = set(schema) - _SCHEMA_TOP_KEYS
    if unknown:
        raise ExecutableFormError(f"unknown schema keywords {sorted(unknown)} — fail closed")
    if schema.get("type", "object") != "object":
        raise ExecutableFormError("schema.type must be 'object' (the act's subject)")
    properties = schema.get("properties", {})
    if not isinstance(properties, dict):
        raise ExecutableFormError("schema.properties must be an object")
    required = schema.get("required", [])
    if not isinstance(required, list) or any(
        not isinstance(name, str) or not name.strip() for name in required
    ):
        raise ExecutableFormError("schema.required must be a list of non-empty field names")
    if len(set(required)) != len(required):
        raise ExecutableFormError("schema.required must not contain duplicate field names")
    additional = schema.get("additionalProperties", True)
    if not isinstance(additional, bool):
        raise ExecutableFormError("schema.additionalProperties must be boolean")
    for pname, pschema in properties.items():
        if not isinstance(pname, str) or not pname.strip():
            raise ExecutableFormError("schema.properties field names must be non-empty strings")
        if not isinstance(pschema, dict):
            raise ExecutableFormError(f"schema.properties[{pname!r}] must be an object")
        bad = set(pschema) - _SCHEMA_PROP_KEYS
        if bad:
            raise ExecutableFormError(
                f"unknown keywords {sorted(bad)} in schema.properties[{pname!r}] — fail closed"
            )
        if "type" in pschema and pschema["type"] not in _SCHEMA_PROP_TYPES:
            raise ExecutableFormError(
                f"schema.properties[{pname!r}].type must be one of {sorted(_SCHEMA_PROP_TYPES)}"
            )
        ptype = pschema.get("type")
        if "enum" in pschema and (
            not isinstance(pschema["enum"], list) or not pschema["enum"]
        ):
            raise ExecutableFormError(
                f"schema.properties[{pname!r}].enum must be a non-empty list"
            )
        for keyword in ("minimum", "maximum"):
            if keyword not in pschema:
                continue
            value = pschema[keyword]
            if (
                not isinstance(value, (int, float))
                or isinstance(value, bool)
                or not math.isfinite(value)
            ):
                raise ExecutableFormError(
                    f"schema.properties[{pname!r}].{keyword} must be a finite number"
                )
            if ptype not in ("integer", "number"):
                raise ExecutableFormError(
                    f"schema.properties[{pname!r}].{keyword} requires numeric type"
                )
        if "minimum" in pschema and "maximum" in pschema \
                and pschema["minimum"] > pschema["maximum"]:
            raise ExecutableFormError(
                f"schema.properties[{pname!r}].minimum must not exceed maximum"
            )
        if "pattern" in pschema:
            if ptype != "string" or not isinstance(pschema["pattern"], str):
                raise ExecutableFormError(
                    f"schema.properties[{pname!r}].pattern requires string type and value"
                )
            try:
                re.compile(pschema["pattern"])
            except re.error as exc:
                raise ExecutableFormError(
                    f"schema.properties[{pname!r}].pattern is invalid: {exc}"
                ) from exc
    quantum = defn.get("quantum")
    if quantum is not None:
        if not isinstance(quantum, dict):
            raise ExecutableFormError("quantum must map field name -> {unit, multipleOf}")
        for fname, q in quantum.items():
            if not isinstance(fname, str) or not fname.strip():
                raise ExecutableFormError("quantum field names must be non-empty strings")
            if not isinstance(q, dict) or set(q) - {"unit", "multipleOf"}:
                raise ExecutableFormError(
                    f"quantum[{fname!r}] must be {{'unit': str, 'multipleOf': int}}"
                )
            if not (q.get("unit") or "").strip():
                raise ExecutableFormError(f"quantum[{fname!r}].unit is required")
            mo = q.get("multipleOf", 1)
            if not isinstance(mo, int) or isinstance(mo, bool) or mo < 1:
                raise ExecutableFormError(
                    f"quantum[{fname!r}].multipleOf must be a positive integer "
                    "(quantized fields are integers in minor units — decidable, no float ties)"
                )
            if fname not in properties or properties[fname].get("type") != "integer":
                raise ExecutableFormError(
                    f"quantum[{fname!r}] requires a declared integer property"
                )


def validate_portable_executable_definition(defn: Any) -> None:
    """Validate the cross-language routed-inference executable subset.

    Legacy ``jsonschema+quantum/1`` permits floats and host-language regex. This
    profile deliberately does not: all numbers are safe integers and patterns are
    absent, so Python, JavaScript, and a zero-import verifier cannot disagree.
    """
    validate_executable_definition(defn)
    schema = defn["schema"]
    for pname, pschema in schema.get("properties", {}).items():
        if pschema.get("type") not in _PORTABLE_PROP_TYPES:
            raise ExecutableFormError(
                f"portable property {pname!r} must declare one of {sorted(_PORTABLE_PROP_TYPES)}"
            )
        if "pattern" in pschema:
            raise ExecutableFormError(
                f"portable property {pname!r} may not use host-language regex"
            )
        for keyword in ("minimum", "maximum", "const"):
            value = pschema.get(keyword)
            if isinstance(value, (int, float)) and not isinstance(value, bool):
                if not isinstance(value, int) or abs(value) > _MAX_SAFE_INTEGER:
                    raise ExecutableFormError(
                        f"portable property {pname!r}.{keyword} must be a safe integer"
                    )
        for value in pschema.get("enum", []):
            if isinstance(value, (int, float)) and not isinstance(value, bool):
                if not isinstance(value, int) or abs(value) > _MAX_SAFE_INTEGER:
                    raise ExecutableFormError(
                        f"portable property {pname!r}.enum values must be safe integers"
                    )


def check_definition(defn: dict, subject: dict) -> tuple[str, list[str]]:
    """Recompute conformance of ``subject`` against a validated executable ``defn``.

    Returns ``(status, reasons)`` with status ``conforms`` / ``violates``. Pure and
    deterministic — no I/O, no globals, no mutation of either argument. Accumulates
    *all* violations (never short-circuits) so the reason list is complete. Assumes
    ``defn`` already passed :func:`validate_executable_definition`."""
    schema, reasons = defn["schema"], []
    props = schema.get("properties") or {}
    for req in schema.get("required") or []:
        if req not in subject:
            reasons.append(f"required field {req!r} absent from action.subject")
    if schema.get("additionalProperties") is False:
        for k in subject:
            if k not in props:
                reasons.append(f"field {k!r} not permitted (additionalProperties: false)")
    for pname, pschema in props.items():
        if pname not in subject:
            continue
        v = subject[pname]
        t = pschema.get("type")
        type_ok = {
            "string": isinstance(v, str),
            "integer": isinstance(v, int) and not isinstance(v, bool),
            "number": isinstance(v, (int, float)) and not isinstance(v, bool),
            "boolean": isinstance(v, bool),
            None: True,
        }[t]
        if not type_ok:
            reasons.append(f"{pname!r} is not of type {t!r}")
            continue
        if "const" in pschema and v != pschema["const"]:
            reasons.append(f"{pname!r} != const {pschema['const']!r}")
        if "enum" in pschema and v not in pschema["enum"]:
            reasons.append(f"{pname!r} not in enum {pschema['enum']!r}")
        if "minimum" in pschema and isinstance(v, (int, float)) and v < pschema["minimum"]:
            reasons.append(f"{pname!r} < minimum {pschema['minimum']}")
        if "maximum" in pschema and isinstance(v, (int, float)) and v > pschema["maximum"]:
            reasons.append(f"{pname!r} > maximum {pschema['maximum']}")
        if "pattern" in pschema and isinstance(v, str):
            import re
            if re.search(pschema["pattern"], v) is None:
                reasons.append(f"{pname!r} does not match pattern {pschema['pattern']!r}")
    for fname, q in (defn.get("quantum") or {}).items():
        if fname not in subject:
            reasons.append(f"quantized field {fname!r} absent from action.subject")
            continue
        v = subject[fname]
        if not isinstance(v, int) or isinstance(v, bool):
            reasons.append(f"quantized field {fname!r} must be an integer in {q['unit']!r}")
        elif v % q.get("multipleOf", 1) != 0:
            reasons.append(f"{fname!r}={v} is not a multiple of {q['multipleOf']} {q['unit']!r}")
    return ("conforms" if not reasons else "violates"), reasons



#
# SPDX-FileCopyrightText: Copyright (c) 2021-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from verifier.config import info_log
from verifier.exceptions import ParsingError

class SpdmMeasurementRequestMessage:
    """ Class representing the SPDM GET_MEASUREMENT request message.
    Following is the expected structure of the MEASUREMENTS request message in DMTF's SPDM 1.1 spec.
    OFFSET   - FIELD                   - SIZE(in bytes)
    0        - SPDMVersion             - 1
    1        - RequestResponseCode     - 1
    2        - Param1                  - 1
    3        - Param2                  - 1
    4        - Nonce                   - 32
    36       - SlotIDParam             - 1
    """
    FieldSize = {
        "SPDMVersion"         : 1,
        "RequestResponseCode" : 1,
        "Param1"              : 1,
        "Param2"              : 1,
        "Nonce"               : 32,
        "SlotIDParam"         : 1,
    }

    def get_spdm_version(self):
        """ Fetches the spdm version field of the object of SpdmMeasurementRequestMessage.

        Returns:
            [bytes]: the spdm version.
        """
        return self.SPDMVersion
    
    def set_spdm_version(self, value):
        """ Sets the spdm version field of the object representing the SPDM GET_MEASUREMENT request.

        Args:
            value (bytes): the spdm version.
        """
        self.SPDMVersion = value

    def get_request_response_code(self):
        """ Fetches the RequestResponseCode field of the object representing the SPDM GET_MEASUREMENT request.

        Returns:
            [bytes]: the RequestResponseCode
        """
        return self.RequestResponseCode
    
    def set_request_response_code(self, value):
        """ Sets the RequestResponseCode field of the object representing the SPDM GET_MEASUREMENT request.

        Args:
            value (bytes): the RequestResponse value.
        """
        self.RequestResponseCode = value

    def get_param1(self):
        """ Fetches the Param1 field of the object representing the SPDM GET_MEASUREMENT request.

        Returns:
            [bytes]: the Param1 value.
        """
        return self.Param1
    
    def set_param1(self, value):
        """ Sets the Param1 field of the object representing the SPDM GET_MEASUREMENT request.

        Args:
            value (bytes): the Param1 value.
        """
        self.Param1 = value

    def get_param2(self):
        """ Fetches the Param2 field of the object representing the SPDM GET_MEASUREMENT request.

        Returns:
            [bytes]: the Param2 value
        """
        return self.Param2

    def set_param2(self, value):
        """ Sets the Param2 field of the object representing the SPDM GET_MEASUREMENT request.

        Args:
            value (bytes): the Param2 value.
        """
        self.Param2 = value

    def get_nonce(self):
        """ Fetches the Nonce field of the object representing the SPDM GET_MEASUREMENT request.

        Returns:
            [bytes]: the nonce value.
        """
        return self.Nonce
    
    def set_nonce(self, value):
        """ Sets the Nonce field value of the object representing the SPDM GET_MEASUREMENT request.

        Args:
            value (bytes): the nonce value.
        """
        self.Nonce = value

    def get_slot_id_param(self):
        """ Fetches the SlotIDParam field value of the object representing the SPDM GET_MEASUREMENT request.

        Returns:
            [bytes]: SlotIDParam value.
        """
        return self.SlotIDParam

    def set_slot_id_param(self, value):
        """ Sets the SlotIDParam field value of the object representing the SPDM GET_MEASUREMENT request.

        Args:
            value (bytes): the SlotIDParam value.
        """
        self.SlotIDParam = value

    def parse(self, request_data):
        """ Parses the raw SPDM GET_MEASUREMENT request message.

        Args:
            request_data (bytes): the raw message data.

        Raises:
            ParsingError: it is raised if there is any incorrect data field length.
        """
        byte_index = 0

        value = request_data[byte_index : byte_index + self.FieldSize['SPDMVersion']]
        self.set_spdm_version(value)
        byte_index = byte_index + self.FieldSize['SPDMVersion']

        value = request_data[byte_index : byte_index + self.FieldSize['RequestResponseCode']]
        self.set_request_response_code(value)
        byte_index = byte_index + self.FieldSize['RequestResponseCode']

        value = request_data[byte_index : byte_index + self.FieldSize['Param1']]
        self.set_param1(value)
        byte_index = byte_index + self.FieldSize['Param1']

        value = request_data[byte_index : byte_index + self.FieldSize['Param2']]
        self.set_param2(value)
        byte_index = byte_index + self.FieldSize['Param2']

        value = request_data[byte_index : byte_index + self.FieldSize['Nonce']]
        self.set_nonce(value)
        byte_index = byte_index + self.FieldSize['Nonce']

        value = request_data[byte_index : byte_index + self.FieldSize['SlotIDParam']]
        self.set_slot_id_param(value)
        byte_index = byte_index + self.FieldSize['SlotIDParam']

        if byte_index != len(request_data):
            err_msg = "Something went wrong during parsing the SPDM GET MEASUREMENT request message."
            info_log.error(err_msg)
            raise ParsingError(err_msg)

    def print_obj(self, logger):
        """ Prints all the field values of the object representing the SPDM GET_MEASUREMENT request.

        Args:
            logger (logging.Logger): the logger object.
        """
        logger.debug("GET MEASUREMENT REQUEST MESSAGE")
        logger.debug(f"SPDMVersion         : {self.SPDMVersion.hex()}")
        logger.debug(f"RequestResponseCode : {self.RequestResponseCode.hex()}")
        logger.debug(f"Param1              : {self.Param1.hex()}")
        logger.debug(f"Param2              : {self.Param2.hex()}")
        logger.debug(f"Nonce               : {self.Nonce.hex()}")
        logger.debug(f"SlotIDParam         : {self.SlotIDParam.hex()}")

    def __init__(self,request_data):
        """ The constructor method for the SpdmMeasurementRequestMessage class representing the SPDM GET_MEASUREMENT
        request message.

        Args:
            request_data (bytes): raw SPDM GET_MEASUREMENT request message.
        """
        assert type(request_data) is bytes

        self.SPDMVersion = None
        self.RequestResponseCode = None
        self.Param1 = None
        self.Param2 = None
        self.Nonce  = None
        self.SlotIDParam = None
        self.parse(request_data)



#!/usr/bin/env python3
"""
AIR v1 vector interoperability test harness.

This script verifies all AIR v1 golden vectors in spec/v1/vectors/ without
depending on the EphemeralML Rust codebase. It performs:

- COSE_Sign1 parsing (CBOR tag 18)
- protected header checks (alg/content_type)
- Ed25519 signature verification over COSE Sig_structure1
- AIR v1 claim validation (Layer 3)
- optional policy checks from vector verify_policy (Layer 4)

Dependencies:
  pip install cbor2 pynacl
"""

from __future__ import annotations

import argparse
import json
import sys
import time
from dataclasses import dataclass
from pathlib import Path
from typing import Any

try:
    import cbor2
except ImportError as exc:  # pragma: no cover - import guard
    print("Missing dependency: cbor2 (pip install cbor2 pynacl)", file=sys.stderr)
    raise SystemExit(2) from exc

try:
    from nacl.exceptions import BadSignatureError
    from nacl.signing import VerifyKey
except ImportError as exc:  # pragma: no cover - import guard
    print("Missing dependency: pynacl (pip install cbor2 pynacl)", file=sys.stderr)
    raise SystemExit(2) from exc


# CWT / EAT keys
CWT_ISS = 1
CWT_IAT = 6
CWT_CTI = 7
EAT_NONCE = 10
EAT_PROFILE = 265

# AIR private claim keys
AIR_MODEL_ID = -65537
AIR_MODEL_VERSION = -65538
AIR_MODEL_HASH = -65539
AIR_REQUEST_HASH = -65540
AIR_RESPONSE_HASH = -65541
AIR_ATTESTATION_DOC_HASH = -65542
AIR_ENCLAVE_MEASUREMENTS = -65543
AIR_POLICY_VERSION = -65544
AIR_SEQUENCE_NUMBER = -65545
AIR_EXECUTION_TIME_MS = -65546
AIR_MEMORY_PEAK_MB = -65547
AIR_SECURITY_MODE = -65548
AIR_MODEL_HASH_SCHEME = -65549

AIR_PROFILE_URI = "https://spec.cyntrisec.com/air/v1"
ALLOWED_MEASUREMENT_TYPES = {"nitro-pcr", "tdx-mrtd-rtmr"}
ALLOWED_MODEL_HASH_SCHEMES = {"sha256-single", "sha256-concat", "sha256-manifest"}

COSE_TAG_SIGN1 = 18
COSE_ALG_KEY = 1
COSE_CONTENT_TYPE_KEY = 3
COSE_ALG_EDDSA = -8
COSE_CONTENT_TYPE_CWT = 61


@dataclass
class VerifyPolicy:
    expected_nonce: bytes | None = None
    expected_model_hash: bytes | None = None
    expected_platform: str | None = None
    expected_model_id: str | None = None
    max_age_secs: int = 0
    clock_skew_secs: int = 0
    require_nonce: bool = False


@dataclass
class VerificationFailure:
    layer: int
    check: str
    code: str
    reason: str


@dataclass
class VerificationResult:
    ok: bool
    failure: VerificationFailure | None = None


class AirVerifyError(Exception):
    def __init__(self, layer: int, check: str, code: str, reason: str) -> None:
        super().__init__(reason)
        self.failure = VerificationFailure(layer=layer, check=check, code=code, reason=reason)


def fail(layer: int, check: str, code: str, reason: str) -> None:
    raise AirVerifyError(layer=layer, check=check, code=code, reason=reason)


def hex_to_bytes(s: str, field: str) -> bytes:
    try:
        return bytes.fromhex(s)
    except ValueError as exc:
        raise ValueError(f"{field} is not valid hex") from exc


def load_json(path: Path) -> dict[str, Any]:
    with path.open("r", encoding="utf-8") as f:
        return json.load(f)


def parse_receipt_cose(receipt_bytes: bytes) -> tuple[bytes, dict[Any, Any], bytes, bytes]:
    try:
        decoded = cbor2.loads(receipt_bytes)
    except Exception as exc:
        fail(1, "COSE", "COSE_DECODE_FAILED", f"failed to decode CBOR: {exc}")

    if not isinstance(decoded, cbor2.CBORTag):
        fail(1, "COSE", "COSE_DECODE_FAILED", "top-level CBOR item is not a tag")
    if decoded.tag != COSE_TAG_SIGN1:
        fail(1, "COSE", "COSE_DECODE_FAILED", f"expected tag 18, got tag {decoded.tag}")

    value = decoded.value
    if not isinstance(value, list) or len(value) != 4:
        fail(1, "COSE", "COSE_DECODE_FAILED", "COSE_Sign1 must be a 4-element array")

    protected, unprotected, payload, signature = value
    if not isinstance(protected, (bytes, bytearray)):
        fail(1, "COSE", "COSE_DECODE_FAILED", "protected header is not bstr")
    if not isinstance(unprotected, dict):
        fail(1, "COSE", "COSE_DECODE_FAILED", "unprotected header is not a map")
    if not isinstance(payload, (bytes, bytearray)) or len(payload) == 0:
        fail(1, "PAYLOAD", "MISSING_PAYLOAD", "payload missing or empty")
    if not isinstance(signature, (bytes, bytearray)):
        fail(2, "SIG", "BAD_SIG_LENGTH", "signature is not bstr")

    return bytes(protected), unprotected, bytes(payload), bytes(signature)


def decode_protected_header(protected_bstr: bytes) -> dict[Any, Any]:
    if protected_bstr == b"":
        return {}
    try:
        hdr = cbor2.loads(protected_bstr)
    except Exception as exc:
        fail(1, "COSE", "COSE_DECODE_FAILED", f"protected header decode failed: {exc}")
    if not isinstance(hdr, dict):
        fail(1, "COSE", "COSE_DECODE_FAILED", "protected header is not a map")
    return hdr


def verify_sig_structure(protected_bstr: bytes, payload: bytes, signature: bytes, public_key: bytes) -> None:
    if len(signature) != 64:
        fail(2, "SIG", "BAD_SIG_LENGTH", f"expected 64-byte Ed25519 signature, got {len(signature)}")
    if len(public_key) != 32:
        fail(2, "SIG", "SIG_FAILED", f"expected 32-byte Ed25519 public key, got {len(public_key)}")

    sig_structure = cbor2.dumps(["Signature1", protected_bstr, b"", payload])
    try:
        VerifyKey(public_key).verify(sig_structure, signature)
    except BadSignatureError:
        fail(2, "SIG", "SIG_FAILED", "Ed25519 verification failed")
    except Exception as exc:
        fail(2, "SIG", "SIG_FAILED", f"Ed25519 verification error: {exc}")


def ensure_type(claims: dict[Any, Any], key: int, expected: type, name: str) -> Any:
    if key not in claims:
        fail(3, name, f"MISSING_CLAIM:{name}", f"missing claim key {key}")
    value = claims[key]
    if not isinstance(value, expected):
        fail(3, name, f"WRONG_TYPE:{name}", f"claim {name} has wrong type: {type(value).__name__}")
    return value


def ensure_bstr_len(value: Any, expected_len: int, layer: int, check: str, code: str, name: str) -> bytes:
    if not isinstance(value, (bytes, bytearray)):
        fail(layer, check, code, f"{name} is not bytes")
    b = bytes(value)
    if len(b) != expected_len:
        fail(layer, check, code, f"{name} length {len(b)} != {expected_len}")
    return b


def decode_and_validate_claims(payload: bytes) -> dict[Any, Any]:
    try:
        claims = cbor2.loads(payload)
    except Exception as exc:
        fail(1, "CLAIMS_DECODE", "PAYLOAD_NOT_MAP", f"claims decode failed: {exc}")
    if not isinstance(claims, dict):
        fail(1, "CLAIMS_DECODE", "PAYLOAD_NOT_MAP", "payload is not a CBOR map")

    # Layer 1-ish (parse/profile): eat_profile
    profile = ensure_type(claims, EAT_PROFILE, str, "EAT_PROFILE")
    if profile != AIR_PROFILE_URI:
        fail(1, "EAT_PROFILE", "WRONG_PROFILE", f"wrong eat_profile: {profile}")

    # Basic required standard claim types
    _ = ensure_type(claims, CWT_ISS, str, "ISS")
    iat = ensure_type(claims, CWT_IAT, int, "IAT")
    if iat < 0:
        fail(3, "IAT", "WRONG_TYPE:IAT", "iat must be unsigned")
    cti = ensure_type(claims, CWT_CTI, (bytes, bytearray), "CTI")
    cti_b = ensure_bstr_len(cti, 16, 3, "CTI", "BAD_CTI_LENGTH", "cti")
    if cti_b == b"\x00" * 16:
        fail(3, "CTI", "BAD_CTI_LENGTH", "cti is all zeros")

    # Optional nonce type
    if EAT_NONCE in claims and not isinstance(claims[EAT_NONCE], (bytes, bytearray)):
        fail(3, "NONCE", "WRONG_TYPE:NONCE", "eat_nonce must be bytes")

    # Required AIR claims
    _ = ensure_type(claims, AIR_MODEL_ID, str, "MODEL_ID")
    _ = ensure_type(claims, AIR_MODEL_VERSION, str, "MODEL_VERSION")
    model_hash = ensure_type(claims, AIR_MODEL_HASH, (bytes, bytearray), "MHASH_PRESENT")
    model_hash_b = ensure_bstr_len(model_hash, 32, 3, "MHASH_PRESENT", "BAD_HASH_LENGTH:model_hash", "model_hash")
    if model_hash_b == b"\x00" * 32:
        fail(3, "MHASH_PRESENT", "ZERO_MODEL_HASH", "model_hash is all zeros")

    ensure_bstr_len(
        ensure_type(claims, AIR_REQUEST_HASH, (bytes, bytearray), "REQUEST_HASH"),
        32,
        3,
        "REQUEST_HASH",
        "BAD_HASH_LENGTH:request_hash",
        "request_hash",
    )
    ensure_bstr_len(
        ensure_type(claims, AIR_RESPONSE_HASH, (bytes, bytearray), "RESPONSE_HASH"),
        32,
        3,
        "RESPONSE_HASH",
        "BAD_HASH_LENGTH:response_hash",
        "response_hash",
    )
    ensure_bstr_len(
        ensure_type(claims, AIR_ATTESTATION_DOC_HASH, (bytes, bytearray), "ATTESTATION_DOC_HASH"),
        32,
        3,
        "ATTESTATION_DOC_HASH",
        "BAD_HASH_LENGTH:attestation_doc_hash",
        "attestation_doc_hash",
    )

    _ = ensure_type(claims, AIR_POLICY_VERSION, str, "POLICY_VERSION")
    _ = ensure_type(claims, AIR_SECURITY_MODE, str, "SECURITY_MODE")
    for key, check in [
        (AIR_SEQUENCE_NUMBER, "SEQ"),
        (AIR_EXECUTION_TIME_MS, "EXEC_MS"),
        (AIR_MEMORY_PEAK_MB, "MEM_MB"),
    ]:
        value = ensure_type(claims, key, int, check)
        if value < 0:
            fail(3, check, f"WRONG_TYPE:{check}", f"{check} must be unsigned")

    # Measurements
    measurements = ensure_type(claims, AIR_ENCLAVE_MEASUREMENTS, dict, "MEAS")
    mtype = measurements.get("measurement_type")
    if not isinstance(mtype, str):
        fail(3, "MTYPE", "UNKNOWN_MTYPE:<non-string>", "measurement_type missing or not a string")
    if mtype not in ALLOWED_MEASUREMENT_TYPES:
        fail(3, "MTYPE", f"UNKNOWN_MTYPE:{mtype}", f"unknown measurement_type: {mtype}")
    for pcr in ("pcr0", "pcr1", "pcr2"):
        if pcr not in measurements:
            fail(3, "MEAS", "BAD_MEASUREMENT_LENGTH", f"missing {pcr}")
        ensure_bstr_len(measurements[pcr], 48, 3, "MEAS", "BAD_MEASUREMENT_LENGTH", pcr)
    if "pcr8" in measurements and measurements["pcr8"] is not None:
        ensure_bstr_len(measurements["pcr8"], 48, 3, "MEAS", "BAD_MEASUREMENT_LENGTH", "pcr8")

    # Optional model_hash_scheme (fail-closed if unknown)
    if AIR_MODEL_HASH_SCHEME in claims:
        scheme = claims[AIR_MODEL_HASH_SCHEME]
        if not isinstance(scheme, str):
            fail(3, "MHASH_SCHEME", "WRONG_TYPE:MHASH_SCHEME", "model_hash_scheme must be a string")
        if scheme not in ALLOWED_MODEL_HASH_SCHEMES:
            fail(3, "MHASH_SCHEME", f"UNKNOWN_MODEL_HASH_SCHEME:{scheme}", f"unknown model_hash_scheme: {scheme}")

    return claims


def apply_policy(claims: dict[Any, Any], policy: VerifyPolicy, now_epoch: int) -> None:
    iat = int(claims[CWT_IAT])
    if policy.max_age_secs > 0:
        if iat > now_epoch + policy.clock_skew_secs:
            fail(4, "FRESH", "TIMESTAMP_FUTURE", "iat is in the future")
        age = max(0, now_epoch - iat)
        if age > policy.max_age_secs:
            fail(4, "FRESH", "TIMESTAMP_STALE", f"receipt age {age}s exceeds max {policy.max_age_secs}s")

    if policy.expected_model_hash is not None:
        actual = bytes(claims[AIR_MODEL_HASH])
        if actual != policy.expected_model_hash:
            fail(4, "MHASH", "MODEL_HASH_MISMATCH", "model_hash does not match policy")

    if policy.expected_model_id is not None:
        actual_id = claims[AIR_MODEL_ID]
        if actual_id != policy.expected_model_id:
            fail(4, "MODEL", "MODEL_ID_MISMATCH", "model_id does not match policy")

    if policy.expected_platform is not None and policy.expected_platform != "any":
        measurements = claims[AIR_ENCLAVE_MEASUREMENTS]
        actual_platform = measurements.get("measurement_type")
        if actual_platform != policy.expected_platform:
            fail(4, "PLATFORM", "PLATFORM_MISMATCH", "measurement_type does not match policy")

    has_nonce = EAT_NONCE in claims
    if policy.require_nonce and not has_nonce:
        fail(4, "NONCE", "NONCE_MISSING", "eat_nonce required but absent")
    if policy.expected_nonce is not None:
        if not has_nonce:
            fail(4, "NONCE", "NONCE_MISSING", "eat_nonce missing")
        actual_nonce = bytes(claims[EAT_NONCE])
        if actual_nonce != policy.expected_nonce:
            fail(4, "NONCE", "NONCE_MISMATCH", "eat_nonce does not match policy")


def build_policy(policy_obj: dict[str, Any] | None) -> VerifyPolicy:
    if not policy_obj:
        return VerifyPolicy()
    policy = VerifyPolicy()
    if "expected_nonce_hex" in policy_obj:
        policy.expected_nonce = hex_to_bytes(policy_obj["expected_nonce_hex"], "expected_nonce_hex")
    if "expected_model_hash_hex" in policy_obj:
        policy.expected_model_hash = hex_to_bytes(policy_obj["expected_model_hash_hex"], "expected_model_hash_hex")
    if "expected_platform" in policy_obj:
        policy.expected_platform = str(policy_obj["expected_platform"])
    if "expected_model_id" in policy_obj:
        policy.expected_model_id = str(policy_obj["expected_model_id"])
    if "max_age_secs" in policy_obj:
        policy.max_age_secs = int(policy_obj["max_age_secs"])
    if "clock_skew_secs" in policy_obj:
        policy.clock_skew_secs = int(policy_obj["clock_skew_secs"])
    if "require_nonce" in policy_obj:
        policy.require_nonce = bool(policy_obj["require_nonce"])
    return policy


def verify_vector(vec: dict[str, Any], now_epoch: int) -> VerificationResult:
    try:
        receipt_bytes = hex_to_bytes(vec["receipt_hex"], "receipt_hex")
        pubkey_field = "public_key_hex" if "public_key_hex" in vec else "wrong_public_key_hex"
        if pubkey_field not in vec:
            raise KeyError("public_key_hex")
        public_key = hex_to_bytes(vec[pubkey_field], pubkey_field)
        policy = build_policy(vec.get("verify_policy"))

        protected_bstr, _uhdr, payload, signature = parse_receipt_cose(receipt_bytes)
        protected = decode_protected_header(protected_bstr)

        alg = protected.get(COSE_ALG_KEY)
        if alg != COSE_ALG_EDDSA:
            fail(1, "ALG", "BAD_ALG", f"alg {alg!r} != -8")

        content_type = protected.get(COSE_CONTENT_TYPE_KEY)
        if content_type != COSE_CONTENT_TYPE_CWT:
            fail(1, "CONTENT_TYPE", "BAD_CONTENT_TYPE", f"content_type {content_type!r} != 61")

        if "payload_hex" in vec:
            expected_payload = hex_to_bytes(vec["payload_hex"], "payload_hex")
            if payload != expected_payload:
                fail(1, "PAYLOAD", "PAYLOAD_NOT_MAP", "payload bytes do not match payload_hex fixture")

        claims = decode_and_validate_claims(payload)
        verify_sig_structure(protected_bstr, payload, signature, public_key)
        apply_policy(claims, policy, now_epoch)
        return VerificationResult(ok=True)
    except AirVerifyError as exc:
        return VerificationResult(ok=False, failure=exc.failure)


def iter_vector_files(vectors_dir: Path, selected_name: str | None) -> list[Path]:
    files = sorted((vectors_dir / "valid").glob("*.json")) + sorted((vectors_dir / "invalid").glob("*.json"))
    if selected_name is None:
        return files
    filtered = [p for p in files if p.stem == selected_name or p.name == selected_name]
    if not filtered:
        raise FileNotFoundError(f"vector not found: {selected_name}")
    return filtered


def evaluate_vector(path: Path, now_epoch: int, verbose: bool) -> bool:
    vec = load_json(path)
    result = verify_vector(vec, now_epoch)
    name = vec.get("name", path.stem)

    if "expected_failure" in vec:
        expected = vec["expected_failure"]
        if result.ok:
            print(f"FAIL {name}: expected failure {expected['code']}, got PASS")
            return False
        actual = result.failure
        assert actual is not None
        ok = (
            actual.code == expected.get("code")
            and actual.check == expected.get("check")
            and actual.layer == int(expected.get("layer"))
        )
        if ok:
            print(f"PASS {name}: expected failure {actual.code} (L{actual.layer}/{actual.check})")
            if verbose:
                print(f"  reason: {actual.reason}")
            return True
        print(
            f"FAIL {name}: expected {expected.get('code')} (L{expected.get('layer')}/{expected.get('check')}), "
            f"got {actual.code} (L{actual.layer}/{actual.check})"
        )
        if verbose:
            print(f"  reason: {actual.reason}")
        return False

    # Valid vector
    if result.ok:
        print(f"PASS {name}: valid vector accepted")
        return True
    assert result.failure is not None
    print(
        f"FAIL {name}: expected PASS, got {result.failure.code} "
        f"(L{result.failure.layer}/{result.failure.check})"
    )
    if verbose:
        print(f"  reason: {result.failure.reason}")
    return False


def main(argv: list[str]) -> int:
    script_dir = Path(__file__).resolve().parent
    default_vectors = (script_dir.parent / "vectors").resolve()

    parser = argparse.ArgumentParser(description="Run AIR v1 vector interoperability checks")
    parser.add_argument("--vectors-dir", type=Path, default=default_vectors, help="Path to spec/v1/vectors")
    parser.add_argument("--vector", help="Run a single vector by filename or stem")
    parser.add_argument("--now-epoch", type=int, default=int(time.time()), help="Unix time for freshness checks")
    parser.add_argument("--verbose", action="store_true", help="Print failure reasons")
    args = parser.parse_args(argv)

    try:
        files = iter_vector_files(args.vectors_dir, args.vector)
    except Exception as exc:
        print(f"Error: {exc}", file=sys.stderr)
        return 2

    if not files:
        print(f"No vector files found under {args.vectors_dir}", file=sys.stderr)
        return 2

    print(f"AIR v1 interop test: {len(files)} vector(s) from {args.vectors_dir}")
    print(f"now_epoch={args.now_epoch}")

    passed = 0
    for path in files:
        if evaluate_vector(path, args.now_epoch, args.verbose):
            passed += 1

    total = len(files)
    failed = total - passed
    print(f"\nSummary: {passed}/{total} passed, {failed} failed")
    return 0 if failed == 0 else 1


if __name__ == "__main__":
    raise SystemExit(main(sys.argv[1:]))
